from __future__ import absolute_import
from __future__ import print_function
import sys
import os

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))))

from veriloggen import *

def mkLed():
    m = Module('blinkled')
    clk = m.Input('CLK')
    rst = m.Input('RST')
    x = m.Input('x', 32)
    vx = m.Input('vx')
    rx = m.Output('rx')
    y = m.Output('y', 32)
    vy = m.Output('vy')
    ry = m.Input('ry')
    
    df = Pipeline(m, 'df', clk, rst)
    
    px = df.input(x, valid=vx, ready=rx)
    py = df(px + 1)
    py.output(y, valid=vy, ready=ry)
    
    df.make_always()

    try:
        df.draw_graph()
    except:
        print('Dataflow graph could not be generated.', file=sys.stderr)

    return m

def mkTest(numports=8):
    m = Module('test')

    # target instance
    led = mkLed()
    
    # copy paras and ports
    params = m.copy_params(led)
    ports = m.copy_sim_ports(led)

    clk = ports['CLK']
    rst = ports['RST']
    x = ports['x']
    vx = ports['vx']
    rx = ports['rx']
    y = ports['y']
    vy = ports['vy']
    ry = ports['ry']
    
    uut = m.Instance(led, 'uut',
                     params=m.connect_params(led),
                     ports=m.connect_ports(led))

    reset_done = m.Reg('reset_done', initval=0)
    
    reset_stmt = []
    reset_stmt.append( reset_done(0) )
    reset_stmt.append( x(0) )
    reset_stmt.append( vx(0) )
    reset_stmt.append( ry(0) )
    
    simulation.setup_waveform(m, uut)
    simulation.setup_clock(m, clk, hperiod=5)
    init = simulation.setup_reset(m, rst, reset_stmt, period=100)

    nclk = simulation.next_clock
    
    init.add(
        Delay(1000),
        reset_done(1),
        nclk(clk),
        Delay(10000),
        Systask('finish'),
    )

    x_count = m.TmpReg(32, initval=0)
    y_count = m.TmpReg(32, initval=0)

    
    xfsm = FSM(m, 'xfsm', clk, rst)
    xfsm.add(vx(0))
    xfsm.goto_next(cond=reset_done)
    for _ in range(10):
        xfsm.goto_next()
    xfsm.add(vx(1))
    xfsm.goto_next()
    xfsm.add(x.inc(), cond=rx)
    xfsm.add(x_count.inc(), cond=rx)
    xfsm.goto_next(cond=AndList(x_count==5, rx))
    xfsm.add(vx(0))
    for _ in range(10):
        xfsm.goto_next()
    xfsm.add(vx(1))
    xfsm.add(x.inc(), cond=rx)
    xfsm.add(x_count.inc(), cond=rx)
    xfsm.goto_next(cond=AndList(x_count==10, rx))
    xfsm.add(vx(0))
    xfsm.make_always()
    
    
    yfsm = FSM(m, 'yfsm', clk, rst)
    yfsm.add(ry(0))
    yfsm.goto_next(cond=reset_done)
    yfsm.goto_next()
    yinit= yfsm.current
    yfsm.add(ry(1), cond=vy)
    yfsm.goto_next(cond=vy)
    for i in range(10):
        yfsm.add(ry(0))
        yfsm.goto_next()
    yfsm.goto(yinit)
    yfsm.make_always()

    
    m.Always(Posedge(clk))(
        If(reset_done)(
            If(AndList(vx, rx))(
                Systask('display', 'x=%d', x)
            ),
            If(AndList(vy, ry))(
                Systask('display', 'y=%d', y)
            )
        )
    )
    
    return m
    
if __name__ == '__main__':
    test = mkTest()
    verilog = test.to_verilog('tmp.v')
    print(verilog)

    # run simulator (Icarus Verilog)
    sim = simulation.Simulator(test)
    rslt = sim.run() # display=False
    #rslt = sim.run(display=True)
    print(rslt)

    # launch waveform viewer (GTKwave)
    #sim.view_waveform() # background=False
    #sim.view_waveform(background=True)



from __future__ import absolute_import
from __future__ import print_function

import os
import sys

import veriloggen.core.vtypes as vtypes


class SubstDstVisitor(object):

    def generic_visit(self, node):
        raise TypeError("Type %s is not supported." % str(type(node)))

    def visit(self, node):
        if isinstance(node, vtypes._Variable):
            return self.visit__Variable(node)

        visitor = getattr(
            self, 'visit_' + node.__class__.__name__, self.generic_visit)
        return visitor(node)

    def visit__Variable(self, node):
        return [node]

    def visit_list(self, node):
        ret = []
        for n in node:
            ret.extend(self.visit(n))
        return ret

    def visit_tuple(self, node):
        ret = []
        for n in node:
            ret.extend(self.visit(n))
        return ret

    def visit_If(self, node):
        true_statement = self.visit(node.true_statement)
        false_statement = (self.visit(node.false_statement)
                           if node.false_statement is not None else [])
        return true_statement + false_statement

    def visit_For(self, node):
        statement = self.visit(node.statement)
        return statement

    def visit_While(self, node):
        statement = self.visit(node.statement)
        return statement

    def visit_Pointer(self, node):
        return [node]

    def visit_Slice(self, node):
        return [node]

    def visit_Cat(self, node):
        return [node]

    def visit_Subst(self, node):
        return self.visit(node.left)

    def visit_SingleStatement(self, node):
        return []



from __future__ import absolute_import
from __future__ import print_function
import sys
import os

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))))

from veriloggen import *
import veriloggen.resolver.resolver as resolver

def mkSubLed():
    m = Module('sub_blinkled')

    width = m.Parameter('WIDTH', 8)
    inc = m.Parameter('INC', 1)
    
    clk = m.Input('CLK')
    rst = m.Input('RST')
    led = m.OutputReg('LED', width)
    
    count = m.Reg('count', width + 10)

    m.Always(Posedge(clk))(
        If(rst)(
            count(0)
        ).Else(
            If(count == 1023)(
                count(0)
            ).Else(
                count(count + inc)
            )
        ))
    
    m.Always(Posedge(clk))(
        If(rst)(
            led( 0 )
        ).Else(
            If(count == 1023)(
                led(led + inc)
            )
        ))

    return m

def mkOrigLed():
    m = Module('blinkled')
    sub = mkSubLed()
    
    width = m.Parameter('TOP_WIDTH', 16)
    inc = m.Parameter('TOP_INC', 1)
    
    clk = m.Input('CLK')
    rst = m.Input('RST')
    led0 = m.Output('LED0', width)
    led1 = m.Output('LED1', width)
    
    m.Instance(sub, 'inst_sub_blinkled_0',
               params=[('WIDTH', width), ('INC', inc + 1)],
               ports=[('CLK', clk), ('RST', rst), ('LED', led0)])
    
    m.Instance(sub, 'inst_sub_blinkled_1',
               params=[('WIDTH', width), ('INC', inc + 2)],
               ports=[('CLK', clk), ('RST', rst), ('LED', led1)])
    
    return m

def mkLed():
    led = mkOrigLed()
    return resolver.resolve(led)

if __name__ == '__main__':
    led = mkLed()
    verilog = led.to_verilog()
    print(verilog)



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import types_axi_axi_to_ram

expected_verilog = """
module test;

  reg CLK;
  reg RST;
  wire [32-1:0] myaxi_awaddr;
  wire [8-1:0] myaxi_awlen;
  wire myaxi_awvalid;
  reg myaxi_awready;
  wire [32-1:0] myaxi_wdata;
  wire [4-1:0] myaxi_wstrb;
  wire myaxi_wlast;
  wire myaxi_wvalid;
  reg myaxi_wready;
  wire [32-1:0] myaxi_araddr;
  wire [8-1:0] myaxi_arlen;
  wire myaxi_arvalid;
  reg myaxi_arready;
  reg [32-1:0] myaxi_rdata;
  reg myaxi_rlast;
  reg myaxi_rvalid;
  wire myaxi_rready;
  wire _tmp_0;
  assign _tmp_0 = 1;

  always @(*) begin
    myaxi_awready <= _tmp_0;
  end

  wire _tmp_1;
  assign _tmp_1 = 1;

  always @(*) begin
    myaxi_wready <= _tmp_1;
  end

  reg [32-1:0] raddr;
  localparam raddr_init = 0;
  reg [32-1:0] _arlen;
  reg [32-1:0] _d1_raddr;
  reg _raddr_cond_3_0_1;
  localparam raddr_1 = 1;
  localparam raddr_2 = 2;
  localparam raddr_3 = 3;
  localparam raddr_4 = 4;
  localparam raddr_5 = 5;
  localparam raddr_6 = 6;
  localparam raddr_7 = 7;
  localparam raddr_8 = 8;
  localparam raddr_9 = 9;
  localparam raddr_10 = 10;

  always @(posedge CLK) begin
    if(RST) begin
      raddr <= raddr_init;
      _d1_raddr <= raddr_init;
      _arlen <= 0;
      _raddr_cond_3_0_1 <= 0;
    end else begin
      _d1_raddr <= raddr;
      case(_d1_raddr)
        raddr_3: begin
          if(_raddr_cond_3_0_1) begin
            myaxi_rvalid <= 0;
            myaxi_rlast <= 0;
          end 
        end
      endcase
      case(raddr)
        raddr_init: begin
          myaxi_arready <= 0;
          myaxi_rdata <= -1;
          myaxi_rvalid <= 0;
          myaxi_rlast <= 0;
          if(myaxi_arvalid) begin
            raddr <= raddr_1;
          end 
        end
        raddr_1: begin
          if(myaxi_arvalid) begin
            myaxi_arready <= 1;
            myaxi_rdata <= myaxi_araddr - 1;
          end 
          raddr <= raddr_2;
        end
        raddr_2: begin
          myaxi_arready <= 0;
          _arlen <= myaxi_arlen;
          raddr <= raddr_3;
        end
        raddr_3: begin
          if((myaxi_rready || !myaxi_rvalid) && !myaxi_rlast) begin
            myaxi_rdata <= myaxi_rdata + 1;
            myaxi_rvalid <= 1;
            myaxi_rlast <= 0;
            _arlen <= _arlen - 1;
          end 
          if((myaxi_rready || !myaxi_rvalid) && !myaxi_rlast && (_arlen == 0)) begin
            myaxi_rlast <= 1;
          end 
          _raddr_cond_3_0_1 <= 1;
          raddr <= raddr_4;
        end
        raddr_4: begin
          if(myaxi_rvalid && !myaxi_rready) begin
            myaxi_rvalid <= myaxi_rvalid;
            myaxi_rlast <= myaxi_rlast;
          end 
          if(myaxi_rvalid && myaxi_rready) begin
            raddr <= raddr_5;
          end 
        end
        raddr_5: begin
          raddr <= raddr_6;
        end
        raddr_6: begin
          raddr <= raddr_7;
        end
        raddr_7: begin
          raddr <= raddr_8;
        end
        raddr_8: begin
          if((_arlen + 1 & 255) != 0) begin
            raddr <= raddr_3;
          end 
          if((_arlen + 1 & 255) == 0) begin
            raddr <= raddr_9;
          end 
        end
        raddr_9: begin
          raddr <= raddr_10;
        end
        raddr_10: begin
          raddr <= raddr_init;
        end
      endcase
    end
  end


  main
  uut
  (
    .CLK(CLK),
    .RST(RST),
    .myaxi_awaddr(myaxi_awaddr),
    .myaxi_awlen(myaxi_awlen),
    .myaxi_awvalid(myaxi_awvalid),
    .myaxi_awready(myaxi_awready),
    .myaxi_wdata(myaxi_wdata),
    .myaxi_wstrb(myaxi_wstrb),
    .myaxi_wlast(myaxi_wlast),
    .myaxi_wvalid(myaxi_wvalid),
    .myaxi_wready(myaxi_wready),
    .myaxi_araddr(myaxi_araddr),
    .myaxi_arlen(myaxi_arlen),
    .myaxi_arvalid(myaxi_arvalid),
    .myaxi_arready(myaxi_arready),
    .myaxi_rdata(myaxi_rdata),
    .myaxi_rlast(myaxi_rlast),
    .myaxi_rvalid(myaxi_rvalid),
    .myaxi_rready(myaxi_rready)
  );


  initial begin
    $dumpfile("uut.vcd");
    $dumpvars(0, uut, CLK, RST, myaxi_awaddr, myaxi_awlen, myaxi_awvalid, myaxi_awready, myaxi_wdata, myaxi_wstrb, myaxi_wlast, myaxi_wvalid, myaxi_wready, myaxi_araddr, myaxi_arlen, myaxi_arvalid, myaxi_arready, myaxi_rdata, myaxi_rlast, myaxi_rvalid, myaxi_rready, _tmp_0, _tmp_1, raddr, _arlen, _d1_raddr, _raddr_cond_3_0_1);
  end


  initial begin
    CLK = 0;
    forever begin
      #5 CLK = !CLK;
    end
  end


  initial begin
    RST = 0;
    raddr = raddr_init;
    _arlen = 0;
    _d1_raddr = raddr_init;
    _raddr_cond_3_0_1 = 0;
    #100;
    RST = 1;
    #100;
    RST = 0;
    #100000;
    $finish;
  end


endmodule



module main
(
  input CLK,
  input RST,
  output reg [32-1:0] myaxi_awaddr,
  output reg [8-1:0] myaxi_awlen,
  output reg myaxi_awvalid,
  input myaxi_awready,
  output reg [32-1:0] myaxi_wdata,
  output reg [4-1:0] myaxi_wstrb,
  output reg myaxi_wlast,
  output reg myaxi_wvalid,
  input myaxi_wready,
  output reg [32-1:0] myaxi_araddr,
  output reg [8-1:0] myaxi_arlen,
  output reg myaxi_arvalid,
  input myaxi_arready,
  input [32-1:0] myaxi_rdata,
  input myaxi_rlast,
  input myaxi_rvalid,
  output myaxi_rready
);

  reg [10-1:0] myram_0_addr;
  wire [32-1:0] myram_0_rdata;
  reg [32-1:0] myram_0_wdata;
  reg myram_0_wenable;

  myram
  inst_myram
  (
    .CLK(CLK),
    .myram_0_addr(myram_0_addr),
    .myram_0_rdata(myram_0_rdata),
    .myram_0_wdata(myram_0_wdata),
    .myram_0_wenable(myram_0_wenable)
  );

  reg [32-1:0] fsm;
  localparam fsm_init = 0;
  reg [9-1:0] _tmp_0;
  reg _myaxi_cond_0_1;
  wire _tmp_1;
  wire _tmp_2;
  assign _tmp_1 = 1 && ((_tmp_ready_21 || !_tmp_valid_21) && (myaxi_rvalid && myaxi_rvalid));
  assign _tmp_2 = 1 && ((_tmp_ready_21 || !_tmp_valid_21) && (myaxi_rvalid && myaxi_rvalid));
  assign myaxi_rready = _tmp_1 && _tmp_2;
  reg [8-1:0] _tmp_3;
  reg _tmp_4;
  wire [32-1:0] _tmp_data_5;
  wire _tmp_valid_5;
  wire _tmp_ready_5;
  assign _tmp_ready_5 = (_tmp_3 > 0) && !_tmp_4;
  reg _myram_cond_0_1;
  reg _tmp_6;
  reg _tmp_7;
  wire _tmp_8;
  wire _tmp_9;
  assign _tmp_8 = 1 && _tmp_ready_18;
  assign _tmp_9 = 1 && _tmp_ready_19;
  localparam _tmp_10 = 1;
  wire [_tmp_10-1:0] _tmp_11;
  assign _tmp_11 = (_tmp_8 || !_tmp_6) && (_tmp_9 || !_tmp_7);
  reg [_tmp_10-1:0] __tmp_11_1;
  wire [32-1:0] _tmp_12;
  reg [32-1:0] __tmp_12_1;
  assign _tmp_12 = (__tmp_11_1)? myram_0_rdata : __tmp_12_1;
  reg [8-1:0] _tmp_13;
  reg _tmp_14;
  reg _tmp_15;
  reg _tmp_16;
  reg _tmp_17;
  wire [32-1:0] _tmp_data_18;
  wire _tmp_valid_18;
  wire _tmp_ready_18;
  assign _tmp_ready_18 = 1;
  wire [1-1:0] _tmp_data_19;
  wire _tmp_valid_19;
  wire _tmp_ready_19;
  assign _tmp_ready_19 = 1;
  reg [32-1:0] sum;
  reg _seq_cond_0_1;

  always @(posedge CLK) begin
    if(RST) begin
      myaxi_awaddr <= 0;
      myaxi_awlen <= 0;
      myaxi_awvalid <= 0;
      myaxi_wdata <= 0;
      myaxi_wstrb <= 0;
      myaxi_wvalid <= 0;
      myaxi_wlast <= 0;
      myaxi_araddr <= 0;
      myaxi_arlen <= 0;
      myaxi_arvalid <= 0;
      _tmp_0 <= 0;
      _myaxi_cond_0_1 <= 0;
    end else begin
      if(_myaxi_cond_0_1) begin
        myaxi_arvalid <= 0;
      end 
      myaxi_awaddr <= 0;
      myaxi_awlen <= 0;
      myaxi_awvalid <= 0;
      myaxi_wdata <= 0;
      myaxi_wstrb <= 0;
      myaxi_wvalid <= 0;
      myaxi_wlast <= 0;
      if((fsm == 0) && ((myaxi_arready || !myaxi_arvalid) && (_tmp_0 == 0))) begin
        myaxi_araddr <= 1024;
        myaxi_arlen <= 63;
        myaxi_arvalid <= 1;
        _tmp_0 <= 64;
      end 
      _myaxi_cond_0_1 <= 1;
      if(myaxi_arvalid && !myaxi_arready) begin
        myaxi_arvalid <= myaxi_arvalid;
      end 
      if(myaxi_rready && myaxi_rvalid && (_tmp_0 > 0)) begin
        _tmp_0 <= _tmp_0 - 1;
      end 
    end
  end


  always @(posedge CLK) begin
    if(RST) begin
      myram_0_addr <= 0;
      _tmp_3 <= 0;
      myram_0_wdata <= 0;
      myram_0_wenable <= 0;
      _tmp_4 <= 0;
      _myram_cond_0_1 <= 0;
      __tmp_11_1 <= 0;
      __tmp_12_1 <= 0;
      _tmp_17 <= 0;
      _tmp_6 <= 0;
      _tmp_7 <= 0;
      _tmp_15 <= 0;
      _tmp_16 <= 0;
      _tmp_14 <= 0;
      _tmp_13 <= 0;
    end else begin
      if(_myram_cond_0_1) begin
        myram_0_wenable <= 0;
        _tmp_4 <= 0;
      end 
      if((fsm == 1) && (_tmp_3 == 0)) begin
        myram_0_addr <= -1;
        _tmp_3 <= 64;
      end 
      if(_tmp_valid_5 && ((_tmp_3 > 0) && !_tmp_4) && (_tmp_3 > 0)) begin
        myram_0_addr <= myram_0_addr + 1;
        myram_0_wdata <= _tmp_data_5;
        myram_0_wenable <= 1;
        _tmp_3 <= _tmp_3 - 1;
      end 
      if(_tmp_valid_5 && ((_tmp_3 > 0) && !_tmp_4) && (_tmp_3 == 1)) begin
        _tmp_4 <= 1;
      end 
      _myram_cond_0_1 <= 1;
      __tmp_11_1 <= _tmp_11;
      __tmp_12_1 <= _tmp_12;
      if((_tmp_8 || !_tmp_6) && (_tmp_9 || !_tmp_7) && _tmp_15) begin
        _tmp_17 <= 0;
        _tmp_6 <= 0;
        _tmp_7 <= 0;
        _tmp_15 <= 0;
      end 
      if((_tmp_8 || !_tmp_6) && (_tmp_9 || !_tmp_7) && _tmp_14) begin
        _tmp_6 <= 1;
        _tmp_7 <= 1;
        _tmp_17 <= _tmp_16;
        _tmp_16 <= 0;
        _tmp_14 <= 0;
        _tmp_15 <= 1;
      end 
      if((fsm == 3) && (_tmp_13 == 0) && !_tmp_16 && !_tmp_17) begin
        myram_0_addr <= 0;
        _tmp_13 <= 63;
        _tmp_14 <= 1;
      end 
      if((_tmp_8 || !_tmp_6) && (_tmp_9 || !_tmp_7) && (_tmp_13 > 0)) begin
        myram_0_addr <= myram_0_addr + 1;
        _tmp_13 <= _tmp_13 - 1;
        _tmp_14 <= 1;
        _tmp_16 <= 0;
      end 
      if((_tmp_8 || !_tmp_6) && (_tmp_9 || !_tmp_7) && (_tmp_13 == 1)) begin
        _tmp_16 <= 1;
      end 
    end
  end

  assign _tmp_data_18 = _tmp_12;
  assign _tmp_valid_18 = _tmp_6;
  assign _tmp_data_19 = _tmp_17;
  assign _tmp_valid_19 = _tmp_7;
  reg [1-1:0] _tmp_data_20;
  reg [32-1:0] _tmp_data_21;
  reg _tmp_valid_21;
  wire _tmp_ready_21;
  assign _tmp_data_5 = _tmp_data_21;
  assign _tmp_valid_5 = _tmp_valid_21;
  assign _tmp_ready_21 = _tmp_ready_5;

  always @(posedge CLK) begin
    if(RST) begin
      _tmp_data_20 <= 0;
      _tmp_data_21 <= 1'd0;
      _tmp_valid_21 <= 0;
    end else begin
      if(myaxi_rvalid && _tmp_2) begin
        _tmp_data_20 <= myaxi_rlast;
      end 
      if((_tmp_ready_21 || !_tmp_valid_21) && (_tmp_1 && _tmp_2) && (myaxi_rvalid && myaxi_rvalid)) begin
        _tmp_data_21 <= _tmp_data_21 + myaxi_rdata;
      end 
      if(_tmp_valid_21 && _tmp_ready_21) begin
        _tmp_valid_21 <= 0;
      end 
      if((_tmp_ready_21 || !_tmp_valid_21) && (_tmp_1 && _tmp_2)) begin
        _tmp_valid_21 <= myaxi_rvalid && myaxi_rvalid;
      end 
      if((_tmp_ready_21 || !_tmp_valid_21) && (_tmp_1 && _tmp_2) && (myaxi_rvalid && myaxi_rvalid) && _tmp_data_20) begin
        _tmp_data_21 <= 1'd0 + myaxi_rdata;
      end 
    end
  end

  localparam fsm_1 = 1;
  localparam fsm_2 = 2;
  localparam fsm_3 = 3;
  localparam fsm_4 = 4;
  localparam fsm_5 = 5;

  always @(posedge CLK) begin
    if(RST) begin
      fsm <= fsm_init;
    end else begin
      case(fsm)
        fsm_init: begin
          if(myaxi_arready || !myaxi_arvalid) begin
            fsm <= fsm_1;
          end 
        end
        fsm_1: begin
          fsm <= fsm_2;
        end
        fsm_2: begin
          if(_tmp_4) begin
            fsm <= fsm_3;
          end 
        end
        fsm_3: begin
          fsm <= fsm_4;
        end
        fsm_4: begin
          if(_tmp_17) begin
            fsm <= fsm_5;
          end 
        end
      endcase
    end
  end


  always @(posedge CLK) begin
    if(RST) begin
      sum <= 0;
      _seq_cond_0_1 <= 0;
    end else begin
      if(_seq_cond_0_1) begin
        $display("sum=%d expected_sum=%d", sum, 2173600);
      end 
      if(_tmp_valid_18) begin
        sum <= sum + _tmp_data_18;
      end 
      _seq_cond_0_1 <= _tmp_valid_18 && (_tmp_data_19 == 1);
    end
  end


endmodule



module myram
(
  input CLK,
  input [10-1:0] myram_0_addr,
  output [32-1:0] myram_0_rdata,
  input [32-1:0] myram_0_wdata,
  input myram_0_wenable
);

  reg [10-1:0] myram_0_daddr;
  reg [32-1:0] mem [0:1024-1];

  always @(posedge CLK) begin
    if(myram_0_wenable) begin
      mem[myram_0_addr] <= myram_0_wdata;
    end 
    myram_0_daddr <= myram_0_addr;
  end

  assign myram_0_rdata = mem[myram_0_daddr];

endmodule
"""


def test():
    veriloggen.reset()
    test_module = types_axi_axi_to_ram.mkTest()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import dataflow_multiple_manager

expected_verilog = """
module test;

  reg CLK;
  reg RST;

  main
  uut
  (
    .CLK(CLK),
    .RST(RST)
  );

  reg reset_done;

  initial begin
    $dumpfile("uut.vcd");
    $dumpvars(0, uut);
  end


  initial begin
    CLK = 0;
    forever begin
      #5 CLK = !CLK;
    end
  end


  initial begin
    RST = 0;
    reset_done = 0;
    #100;
    RST = 1;
    #100;
    RST = 0;
    #1000;
    reset_done = 1;
    @(posedge CLK);
    #1;
    #10000;
    $finish;
  end


endmodule



module main
(
  input CLK,
  input RST
);

  reg [32-1:0] _tmp_data_0;
  reg _tmp_valid_0;
  wire _tmp_ready_0;
  assign _tmp_ready_0 = (_tmp_ready_3 || !_tmp_valid_3) && (_tmp_valid_0 && _tmp_valid_1) && ((_tmp_ready_5 || !_tmp_valid_5) && (_tmp_valid_2 && _tmp_valid_0)) && ((_tmp_ready_7 || !_tmp_valid_7) && _tmp_valid_0);
  reg [32-1:0] _tmp_data_1;
  reg _tmp_valid_1;
  wire _tmp_ready_1;
  assign _tmp_ready_1 = (_tmp_ready_3 || !_tmp_valid_3) && (_tmp_valid_0 && _tmp_valid_1) && ((_tmp_ready_4 || !_tmp_valid_4) && (_tmp_valid_1 && _tmp_valid_2)) && ((_tmp_ready_8 || !_tmp_valid_8) && _tmp_valid_1);
  reg [32-1:0] _tmp_data_2;
  reg _tmp_valid_2;
  wire _tmp_ready_2;
  assign _tmp_ready_2 = (_tmp_ready_4 || !_tmp_valid_4) && (_tmp_valid_1 && _tmp_valid_2) && ((_tmp_ready_5 || !_tmp_valid_5) && (_tmp_valid_2 && _tmp_valid_0)) && ((_tmp_ready_6 || !_tmp_valid_6) && _tmp_valid_2);
  reg [32-1:0] _tmp_data_3;
  reg _tmp_valid_3;
  wire _tmp_ready_3;
  assign _tmp_ready_3 = (_tmp_ready_9 || !_tmp_valid_9) && (_tmp_valid_3 && _tmp_valid_6);
  reg [32-1:0] _tmp_data_4;
  reg _tmp_valid_4;
  wire _tmp_ready_4;
  assign _tmp_ready_4 = (_tmp_ready_10 || !_tmp_valid_10) && (_tmp_valid_4 && _tmp_valid_7);
  reg [32-1:0] _tmp_data_5;
  reg _tmp_valid_5;
  wire _tmp_ready_5;
  assign _tmp_ready_5 = (_tmp_ready_11 || !_tmp_valid_11) && (_tmp_valid_5 && _tmp_valid_8);
  reg [32-1:0] _tmp_data_6;
  reg _tmp_valid_6;
  wire _tmp_ready_6;
  assign _tmp_ready_6 = (_tmp_ready_9 || !_tmp_valid_9) && (_tmp_valid_3 && _tmp_valid_6);
  reg [32-1:0] _tmp_data_7;
  reg _tmp_valid_7;
  wire _tmp_ready_7;
  assign _tmp_ready_7 = (_tmp_ready_10 || !_tmp_valid_10) && (_tmp_valid_4 && _tmp_valid_7);
  reg [32-1:0] _tmp_data_8;
  reg _tmp_valid_8;
  wire _tmp_ready_8;
  assign _tmp_ready_8 = (_tmp_ready_11 || !_tmp_valid_11) && (_tmp_valid_5 && _tmp_valid_8);
  reg [32-1:0] _tmp_data_9;
  reg _tmp_valid_9;
  wire _tmp_ready_9;
  reg [32-1:0] _tmp_data_10;
  reg _tmp_valid_10;
  wire _tmp_ready_10;
  reg [32-1:0] _tmp_data_11;
  reg _tmp_valid_11;
  wire _tmp_ready_11;
  wire [32-1:0] adata;
  wire avalid;
  assign adata = _tmp_data_9;
  assign avalid = _tmp_valid_9;
  assign _tmp_ready_9 = 1;
  wire [32-1:0] bdata;
  wire bvalid;
  assign bdata = _tmp_data_10;
  assign bvalid = _tmp_valid_10;
  assign _tmp_ready_10 = 1;
  wire [32-1:0] cdata;
  wire cvalid;
  assign cdata = _tmp_data_11;
  assign cvalid = _tmp_valid_11;
  assign _tmp_ready_11 = 1;

  always @(posedge CLK) begin
    if(RST) begin
      _tmp_data_0 <= 1'd0;
      _tmp_valid_0 <= 0;
      _tmp_data_1 <= 1'd0;
      _tmp_valid_1 <= 0;
      _tmp_data_2 <= 1'd0;
      _tmp_valid_2 <= 0;
      _tmp_data_3 <= 0;
      _tmp_valid_3 <= 0;
      _tmp_data_4 <= 0;
      _tmp_valid_4 <= 0;
      _tmp_data_5 <= 0;
      _tmp_valid_5 <= 0;
      _tmp_data_6 <= 0;
      _tmp_valid_6 <= 0;
      _tmp_data_7 <= 0;
      _tmp_valid_7 <= 0;
      _tmp_data_8 <= 0;
      _tmp_valid_8 <= 0;
      _tmp_data_9 <= 0;
      _tmp_valid_9 <= 0;
      _tmp_data_10 <= 0;
      _tmp_valid_10 <= 0;
      _tmp_data_11 <= 0;
      _tmp_valid_11 <= 0;
    end else begin
      if((_tmp_ready_0 || !_tmp_valid_0) && 1 && 1) begin
        _tmp_data_0 <= _tmp_data_0 + 2'd1;
      end 
      if(_tmp_valid_0 && _tmp_ready_0) begin
        _tmp_valid_0 <= 0;
      end 
      if((_tmp_ready_0 || !_tmp_valid_0) && 1) begin
        _tmp_valid_0 <= 1;
      end 
      if((_tmp_ready_1 || !_tmp_valid_1) && 1 && 1) begin
        _tmp_data_1 <= _tmp_data_1 + 2'd1;
      end 
      if(_tmp_valid_1 && _tmp_ready_1) begin
        _tmp_valid_1 <= 0;
      end 
      if((_tmp_ready_1 || !_tmp_valid_1) && 1) begin
        _tmp_valid_1 <= 1;
      end 
      if((_tmp_ready_2 || !_tmp_valid_2) && 1 && 1) begin
        _tmp_data_2 <= _tmp_data_2 + 2'd1;
      end 
      if(_tmp_valid_2 && _tmp_ready_2) begin
        _tmp_valid_2 <= 0;
      end 
      if((_tmp_ready_2 || !_tmp_valid_2) && 1) begin
        _tmp_valid_2 <= 1;
      end 
      if((_tmp_ready_3 || !_tmp_valid_3) && (_tmp_ready_0 && _tmp_ready_1) && (_tmp_valid_0 && _tmp_valid_1)) begin
        _tmp_data_3 <= _tmp_data_0 + _tmp_data_1;
      end 
      if(_tmp_valid_3 && _tmp_ready_3) begin
        _tmp_valid_3 <= 0;
      end 
      if((_tmp_ready_3 || !_tmp_valid_3) && (_tmp_ready_0 && _tmp_ready_1)) begin
        _tmp_valid_3 <= _tmp_valid_0 && _tmp_valid_1;
      end 
      if((_tmp_ready_4 || !_tmp_valid_4) && (_tmp_ready_1 && _tmp_ready_2) && (_tmp_valid_1 && _tmp_valid_2)) begin
        _tmp_data_4 <= _tmp_data_1 + _tmp_data_2;
      end 
      if(_tmp_valid_4 && _tmp_ready_4) begin
        _tmp_valid_4 <= 0;
      end 
      if((_tmp_ready_4 || !_tmp_valid_4) && (_tmp_ready_1 && _tmp_ready_2)) begin
        _tmp_valid_4 <= _tmp_valid_1 && _tmp_valid_2;
      end 
      if((_tmp_ready_5 || !_tmp_valid_5) && (_tmp_ready_2 && _tmp_ready_0) && (_tmp_valid_2 && _tmp_valid_0)) begin
        _tmp_data_5 <= _tmp_data_2 + _tmp_data_0;
      end 
      if(_tmp_valid_5 && _tmp_ready_5) begin
        _tmp_valid_5 <= 0;
      end 
      if((_tmp_ready_5 || !_tmp_valid_5) && (_tmp_ready_2 && _tmp_ready_0)) begin
        _tmp_valid_5 <= _tmp_valid_2 && _tmp_valid_0;
      end 
      if((_tmp_ready_6 || !_tmp_valid_6) && _tmp_ready_2 && _tmp_valid_2) begin
        _tmp_data_6 <= _tmp_data_2;
      end 
      if(_tmp_valid_6 && _tmp_ready_6) begin
        _tmp_valid_6 <= 0;
      end 
      if((_tmp_ready_6 || !_tmp_valid_6) && _tmp_ready_2) begin
        _tmp_valid_6 <= _tmp_valid_2;
      end 
      if((_tmp_ready_7 || !_tmp_valid_7) && _tmp_ready_0 && _tmp_valid_0) begin
        _tmp_data_7 <= _tmp_data_0;
      end 
      if(_tmp_valid_7 && _tmp_ready_7) begin
        _tmp_valid_7 <= 0;
      end 
      if((_tmp_ready_7 || !_tmp_valid_7) && _tmp_ready_0) begin
        _tmp_valid_7 <= _tmp_valid_0;
      end 
      if((_tmp_ready_8 || !_tmp_valid_8) && _tmp_ready_1 && _tmp_valid_1) begin
        _tmp_data_8 <= _tmp_data_1;
      end 
      if(_tmp_valid_8 && _tmp_ready_8) begin
        _tmp_valid_8 <= 0;
      end 
      if((_tmp_ready_8 || !_tmp_valid_8) && _tmp_ready_1) begin
        _tmp_valid_8 <= _tmp_valid_1;
      end 
      if((_tmp_ready_9 || !_tmp_valid_9) && (_tmp_ready_3 && _tmp_ready_6) && (_tmp_valid_3 && _tmp_valid_6)) begin
        _tmp_data_9 <= _tmp_data_3 + _tmp_data_6;
      end 
      if(_tmp_valid_9 && _tmp_ready_9) begin
        _tmp_valid_9 <= 0;
      end 
      if((_tmp_ready_9 || !_tmp_valid_9) && (_tmp_ready_3 && _tmp_ready_6)) begin
        _tmp_valid_9 <= _tmp_valid_3 && _tmp_valid_6;
      end 
      if((_tmp_ready_10 || !_tmp_valid_10) && (_tmp_ready_4 && _tmp_ready_7) && (_tmp_valid_4 && _tmp_valid_7)) begin
        _tmp_data_10 <= _tmp_data_4 + _tmp_data_7;
      end 
      if(_tmp_valid_10 && _tmp_ready_10) begin
        _tmp_valid_10 <= 0;
      end 
      if((_tmp_ready_10 || !_tmp_valid_10) && (_tmp_ready_4 && _tmp_ready_7)) begin
        _tmp_valid_10 <= _tmp_valid_4 && _tmp_valid_7;
      end 
      if((_tmp_ready_11 || !_tmp_valid_11) && (_tmp_ready_5 && _tmp_ready_8) && (_tmp_valid_5 && _tmp_valid_8)) begin
        _tmp_data_11 <= _tmp_data_5 + _tmp_data_8;
      end 
      if(_tmp_valid_11 && _tmp_ready_11) begin
        _tmp_valid_11 <= 0;
      end 
      if((_tmp_ready_11 || !_tmp_valid_11) && (_tmp_ready_5 && _tmp_ready_8)) begin
        _tmp_valid_11 <= _tmp_valid_5 && _tmp_valid_8;
      end 
    end
  end


endmodule
"""

def test():
    veriloggen.reset()
    test_module = dataflow_multiple_manager.mkTest()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import uart

expected_verilog = """
module test;

  reg CLK;
  reg RST;
  reg [8-1:0] din;
  reg enable;
  wire ready;
  wire txd;
  wire [8-1:0] dout;
  wire valid;
  wire rxd;
  assign rxd = txd;
  reg [32-1:0] txfsm;
  localparam txfsm_init = 0;
  reg [32-1:0] _tmp_0;
  localparam txfsm_1 = 1;
  localparam txfsm_2 = 2;
  localparam txfsm_3 = 3;
  localparam txfsm_4 = 4;
  localparam txfsm_5 = 5;
  localparam txfsm_6 = 6;
  localparam txfsm_7 = 7;
  localparam txfsm_8 = 8;

  always @(posedge CLK) begin
    if(RST) begin
      txfsm <= txfsm_init;
    end else begin
      case(txfsm)
        txfsm_init: begin
          if(ready) begin
            din <= 85;
            enable <= 1;
          end 
          if(ready) begin
            txfsm <= txfsm_1;
          end 
        end
        txfsm_1: begin
          enable <= 0;
          txfsm <= txfsm_2;
        end
        txfsm_2: begin
          if(ready) begin
            din <= 51;
            enable <= 1;
          end 
          if(ready) begin
            txfsm <= txfsm_3;
          end 
        end
        txfsm_3: begin
          enable <= 0;
          txfsm <= txfsm_4;
        end
        txfsm_4: begin
          if(ready) begin
            din <= 0;
            enable <= 1;
          end 
          if(ready) begin
            txfsm <= txfsm_5;
          end 
        end
        txfsm_5: begin
          enable <= 0;
          txfsm <= txfsm_6;
        end
        txfsm_6: begin
          if(ready) begin
            din <= 255;
            enable <= 1;
          end 
          if(ready) begin
            txfsm <= txfsm_7;
          end 
        end
        txfsm_7: begin
          enable <= 0;
          txfsm <= txfsm_8;
        end
      endcase
    end
  end

  reg [32-1:0] rxfsm;
  localparam rxfsm_init = 0;
  localparam rxfsm_1 = 1;
  localparam rxfsm_2 = 2;
  localparam rxfsm_3 = 3;
  localparam rxfsm_4 = 4;

  always @(posedge CLK) begin
    if(RST) begin
      rxfsm <= rxfsm_init;
    end else begin
      case(rxfsm)
        rxfsm_init: begin
          if(valid) begin
            $display("din=%02x dout=%02x %s", 8'd85, dout, ((dout == 85)? "OK" : "NG"));
          end 
          if(valid) begin
            rxfsm <= rxfsm_1;
          end 
        end
        rxfsm_1: begin
          if(valid) begin
            $display("din=%02x dout=%02x %s", 8'd51, dout, ((dout == 51)? "OK" : "NG"));
          end 
          if(valid) begin
            rxfsm <= rxfsm_2;
          end 
        end
        rxfsm_2: begin
          if(valid) begin
            $display("din=%02x dout=%02x %s", 8'd0, dout, ((dout == 0)? "OK" : "NG"));
          end 
          if(valid) begin
            rxfsm <= rxfsm_3;
          end 
        end
        rxfsm_3: begin
          if(valid) begin
            $display("din=%02x dout=%02x %s", 8'd255, dout, ((dout == 255)? "OK" : "NG"));
          end 
          if(valid) begin
            rxfsm <= rxfsm_4;
          end 
        end
      endcase
    end
  end


  UartTx
  mtx
  (
    .CLK(CLK),
    .RST(RST),
    .din(din),
    .enable(enable),
    .ready(ready),
    .txd(txd)
  );


  UartRx
  mrx
  (
    .CLK(CLK),
    .RST(RST),
    .rxd(rxd),
    .dout(dout),
    .valid(valid)
  );


  initial begin
    $dumpfile("uut.vcd");
    $dumpvars(0, mtx, mrx, CLK, RST, din, enable, ready, txd, dout, valid, rxd, txfsm, _tmp_0, rxfsm);
  end


  initial begin
    CLK = 0;
    forever begin
      #5 CLK = !CLK;
    end
  end


  initial begin
    RST = 0;
    txfsm = txfsm_init;
    _tmp_0 = 0;
    rxfsm = rxfsm_init;
    #100;
    RST = 1;
    #100;
    RST = 0;
    #100000;
    $finish;
  end


endmodule



module UartTx
(
  input CLK,
  input RST,
  input [8-1:0] din,
  input enable,
  output reg ready,
  output reg txd
);

  reg [32-1:0] fsm;
  localparam fsm_init = 0;
  reg [9-1:0] _tmp_0;
  reg [4-1:0] _tmp_1;
  localparam fsm_1 = 1;
  localparam fsm_2 = 2;
  localparam fsm_3 = 3;
  localparam fsm_4 = 4;
  localparam fsm_5 = 5;
  localparam fsm_6 = 6;
  localparam fsm_7 = 7;
  localparam fsm_8 = 8;
  localparam fsm_9 = 9;
  localparam fsm_10 = 10;
  localparam fsm_11 = 11;

  always @(posedge CLK) begin
    if(RST) begin
      fsm <= fsm_init;
      _tmp_1 <= 0;
      txd <= 1;
      _tmp_0 <= 0;
      ready <= 1;
    end else begin
      case(fsm)
        fsm_init: begin
          _tmp_1 <= 9;
          txd <= 1;
          _tmp_0 <= { din, 1'd0 };
          if(enable) begin
            ready <= 0;
          end 
          if(enable) begin
            fsm <= fsm_1;
          end 
        end
        fsm_1: begin
          if(_tmp_1 > 0) begin
            _tmp_1 <= _tmp_1 - 1;
          end else begin
            txd <= _tmp_0[0];
            _tmp_0 <= { 1'd1, _tmp_0[8:1] };
            _tmp_1 <= 9;
          end
          if(!(_tmp_1 > 0)) begin
            fsm <= fsm_2;
          end 
        end
        fsm_2: begin
          if(_tmp_1 > 0) begin
            _tmp_1 <= _tmp_1 - 1;
          end else begin
            txd <= _tmp_0[0];
            _tmp_0 <= { 1'd1, _tmp_0[8:1] };
            _tmp_1 <= 9;
          end
          if(!(_tmp_1 > 0)) begin
            fsm <= fsm_3;
          end 
        end
        fsm_3: begin
          if(_tmp_1 > 0) begin
            _tmp_1 <= _tmp_1 - 1;
          end else begin
            txd <= _tmp_0[0];
            _tmp_0 <= { 1'd1, _tmp_0[8:1] };
            _tmp_1 <= 9;
          end
          if(!(_tmp_1 > 0)) begin
            fsm <= fsm_4;
          end 
        end
        fsm_4: begin
          if(_tmp_1 > 0) begin
            _tmp_1 <= _tmp_1 - 1;
          end else begin
            txd <= _tmp_0[0];
            _tmp_0 <= { 1'd1, _tmp_0[8:1] };
            _tmp_1 <= 9;
          end
          if(!(_tmp_1 > 0)) begin
            fsm <= fsm_5;
          end 
        end
        fsm_5: begin
          if(_tmp_1 > 0) begin
            _tmp_1 <= _tmp_1 - 1;
          end else begin
            txd <= _tmp_0[0];
            _tmp_0 <= { 1'd1, _tmp_0[8:1] };
            _tmp_1 <= 9;
          end
          if(!(_tmp_1 > 0)) begin
            fsm <= fsm_6;
          end 
        end
        fsm_6: begin
          if(_tmp_1 > 0) begin
            _tmp_1 <= _tmp_1 - 1;
          end else begin
            txd <= _tmp_0[0];
            _tmp_0 <= { 1'd1, _tmp_0[8:1] };
            _tmp_1 <= 9;
          end
          if(!(_tmp_1 > 0)) begin
            fsm <= fsm_7;
          end 
        end
        fsm_7: begin
          if(_tmp_1 > 0) begin
            _tmp_1 <= _tmp_1 - 1;
          end else begin
            txd <= _tmp_0[0];
            _tmp_0 <= { 1'd1, _tmp_0[8:1] };
            _tmp_1 <= 9;
          end
          if(!(_tmp_1 > 0)) begin
            fsm <= fsm_8;
          end 
        end
        fsm_8: begin
          if(_tmp_1 > 0) begin
            _tmp_1 <= _tmp_1 - 1;
          end else begin
            txd <= _tmp_0[0];
            _tmp_0 <= { 1'd1, _tmp_0[8:1] };
            _tmp_1 <= 9;
          end
          if(!(_tmp_1 > 0)) begin
            fsm <= fsm_9;
          end 
        end
        fsm_9: begin
          if(_tmp_1 > 0) begin
            _tmp_1 <= _tmp_1 - 1;
          end else begin
            txd <= _tmp_0[0];
            _tmp_0 <= { 1'd1, _tmp_0[8:1] };
            _tmp_1 <= 9;
          end
          if(!(_tmp_1 > 0)) begin
            fsm <= fsm_10;
          end 
        end
        fsm_10: begin
          if(_tmp_1 > 0) begin
            _tmp_1 <= _tmp_1 - 1;
          end else begin
            txd <= _tmp_0[0];
            _tmp_0 <= { 1'd1, _tmp_0[8:1] };
            _tmp_1 <= 9;
          end
          if(!(_tmp_1 > 0)) begin
            fsm <= fsm_11;
          end 
        end
        fsm_11: begin
          ready <= 1;
          fsm <= fsm_init;
        end
      endcase
    end
  end


endmodule



module UartRx
(
  input CLK,
  input RST,
  input rxd,
  output reg [8-1:0] dout,
  output reg valid
);

  reg [32-1:0] fsm;
  localparam fsm_init = 0;
  reg [9-1:0] _tmp_0;
  reg [4-1:0] _tmp_1;
  localparam fsm_1 = 1;
  localparam fsm_2 = 2;
  localparam fsm_3 = 3;
  localparam fsm_4 = 4;
  localparam fsm_5 = 5;
  localparam fsm_6 = 6;
  localparam fsm_7 = 7;
  localparam fsm_8 = 8;
  localparam fsm_9 = 9;
  localparam fsm_10 = 10;
  localparam fsm_11 = 11;

  always @(posedge CLK) begin
    if(RST) begin
      fsm <= fsm_init;
      valid <= 0;
      _tmp_1 <= 0;
      _tmp_0 <= 0;
      dout <= 0;
    end else begin
      case(fsm)
        fsm_init: begin
          valid <= 0;
          _tmp_1 <= 4;
          _tmp_0 <= { rxd, _tmp_0[8:1] };
          if(rxd == 0) begin
            fsm <= fsm_1;
          end 
        end
        fsm_1: begin
          if(_tmp_1 > 0) begin
            _tmp_1 <= _tmp_1 - 1;
          end else begin
            _tmp_0 <= { rxd, _tmp_0[8:1] };
            _tmp_1 <= 9;
          end
          if((_tmp_1 == 1) && (rxd != 0)) begin
            fsm <= fsm_init;
          end 
          if(!(_tmp_1 > 0)) begin
            fsm <= fsm_2;
          end 
        end
        fsm_2: begin
          if(_tmp_1 > 0) begin
            _tmp_1 <= _tmp_1 - 1;
          end else begin
            _tmp_0 <= { rxd, _tmp_0[8:1] };
            _tmp_1 <= 9;
          end
          if(!(_tmp_1 > 0)) begin
            fsm <= fsm_3;
          end 
        end
        fsm_3: begin
          if(_tmp_1 > 0) begin
            _tmp_1 <= _tmp_1 - 1;
          end else begin
            _tmp_0 <= { rxd, _tmp_0[8:1] };
            _tmp_1 <= 9;
          end
          if(!(_tmp_1 > 0)) begin
            fsm <= fsm_4;
          end 
        end
        fsm_4: begin
          if(_tmp_1 > 0) begin
            _tmp_1 <= _tmp_1 - 1;
          end else begin
            _tmp_0 <= { rxd, _tmp_0[8:1] };
            _tmp_1 <= 9;
          end
          if(!(_tmp_1 > 0)) begin
            fsm <= fsm_5;
          end 
        end
        fsm_5: begin
          if(_tmp_1 > 0) begin
            _tmp_1 <= _tmp_1 - 1;
          end else begin
            _tmp_0 <= { rxd, _tmp_0[8:1] };
            _tmp_1 <= 9;
          end
          if(!(_tmp_1 > 0)) begin
            fsm <= fsm_6;
          end 
        end
        fsm_6: begin
          if(_tmp_1 > 0) begin
            _tmp_1 <= _tmp_1 - 1;
          end else begin
            _tmp_0 <= { rxd, _tmp_0[8:1] };
            _tmp_1 <= 9;
          end
          if(!(_tmp_1 > 0)) begin
            fsm <= fsm_7;
          end 
        end
        fsm_7: begin
          if(_tmp_1 > 0) begin
            _tmp_1 <= _tmp_1 - 1;
          end else begin
            _tmp_0 <= { rxd, _tmp_0[8:1] };
            _tmp_1 <= 9;
          end
          if(!(_tmp_1 > 0)) begin
            fsm <= fsm_8;
          end 
        end
        fsm_8: begin
          if(_tmp_1 > 0) begin
            _tmp_1 <= _tmp_1 - 1;
          end else begin
            _tmp_0 <= { rxd, _tmp_0[8:1] };
            _tmp_1 <= 9;
          end
          if(!(_tmp_1 > 0)) begin
            fsm <= fsm_9;
          end 
        end
        fsm_9: begin
          if(_tmp_1 > 0) begin
            _tmp_1 <= _tmp_1 - 1;
          end else begin
            _tmp_0 <= { rxd, _tmp_0[8:1] };
            _tmp_1 <= 9;
          end
          if(!(_tmp_1 > 0)) begin
            fsm <= fsm_10;
          end 
        end
        fsm_10: begin
          if(_tmp_1 > 0) begin
            _tmp_1 <= _tmp_1 - 1;
          end else begin
            _tmp_0 <= { rxd, _tmp_0[8:1] };
            _tmp_1 <= 9;
          end
          if(!(_tmp_1 > 0)) begin
            fsm <= fsm_11;
          end 
        end
        fsm_11: begin
          valid <= 1;
          dout <= _tmp_0[8:0];
          fsm <= fsm_init;
        end
      endcase
    end
  end


endmodule
"""

def test():
    veriloggen.reset()
    test_module = uart.mkTest()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import types_ram_manager_rtl_dataflow

expected_verilog = """
module test;

  reg CLK;
  reg RST;

  main
  uut
  (
    .CLK(CLK),
    .RST(RST)
  );


  initial begin
    $dumpfile("uut.vcd");
    $dumpvars(0, uut, CLK, RST);
  end


  initial begin
    CLK = 0;
    forever begin
      #5 CLK = !CLK;
    end
  end


  initial begin
    RST = 0;
    #100;
    RST = 1;
    #100;
    RST = 0;
    #100000;
    $finish;
  end


endmodule



module main
(
  input CLK,
  input RST
);

  reg [14-1:0] myram_0_addr;
  wire [32-1:0] myram_0_rdata;
  reg [32-1:0] myram_0_wdata;
  reg myram_0_wenable;
  reg [14-1:0] myram_1_addr;
  wire [32-1:0] myram_1_rdata;
  reg [32-1:0] myram_1_wdata;
  reg myram_1_wenable;

  myram
  inst_myram
  (
    .CLK(CLK),
    .myram_0_addr(myram_0_addr),
    .myram_0_rdata(myram_0_rdata),
    .myram_0_wdata(myram_0_wdata),
    .myram_0_wenable(myram_0_wenable),
    .myram_1_addr(myram_1_addr),
    .myram_1_rdata(myram_1_rdata),
    .myram_1_wdata(myram_1_wdata),
    .myram_1_wenable(myram_1_wenable)
  );

  reg [32-1:0] xfsm;
  localparam xfsm_init = 0;
  reg [32-1:0] xaddr;
  reg [32-1:0] xcount;
  wire [32-1:0] xdata;
  wire xvalid;
  wire xready;
  reg [32-1:0] _tmp_data_0;
  reg _tmp_valid_0;
  wire _tmp_ready_0;
  assign xready = (_tmp_ready_0 || !_tmp_valid_0) && xvalid;
  wire [32-1:0] ydata;
  wire yvalid;
  wire yread;
  assign ydata = _tmp_data_0;
  assign yvalid = _tmp_valid_0;
  assign _tmp_ready_0 = yread;
  reg _myram_cond_0_1;
  reg _tmp_1;
  reg _myram_cond_1_1;
  reg _myram_cond_2_1;
  reg _myram_cond_2_2;
  reg [32-1:0] _tmp_2;
  assign xdata = _tmp_2;
  reg _tmp_3;
  assign xvalid = _tmp_3;
  reg __dataflow_seq_0_cond_0_1;
  reg [32-1:0] _xaddr_1;
  reg [32-1:0] _xaddr_2;
  reg [32-1:0] yfsm;
  localparam yfsm_init = 0;
  reg [32-1:0] yaddr;
  assign yread = yfsm == 1;
  reg _myram_cond_3_1;
  reg _tmp_4;
  reg _myram_cond_4_1;
  reg _myram_cond_5_1;
  reg _myram_cond_5_2;
  reg [32-1:0] _yaddr_1;
  reg [32-1:0] _yaddr_2;

  always @(posedge CLK) begin
    if(RST) begin
      myram_0_addr <= 0;
      myram_0_wdata <= 0;
      myram_0_wenable <= 0;
      _myram_cond_0_1 <= 0;
      _myram_cond_1_1 <= 0;
      _tmp_1 <= 0;
      _myram_cond_2_1 <= 0;
      _myram_cond_2_2 <= 0;
      myram_1_addr <= 0;
      myram_1_wdata <= 0;
      myram_1_wenable <= 0;
      _myram_cond_3_1 <= 0;
      _myram_cond_4_1 <= 0;
      _tmp_4 <= 0;
      _myram_cond_5_1 <= 0;
      _myram_cond_5_2 <= 0;
    end else begin
      if(_myram_cond_2_2) begin
        _tmp_1 <= 0;
      end 
      if(_myram_cond_5_2) begin
        _tmp_4 <= 0;
      end 
      if(_myram_cond_0_1) begin
        myram_0_wenable <= 0;
      end 
      if(_myram_cond_1_1) begin
        _tmp_1 <= 1;
      end 
      _myram_cond_2_2 <= _myram_cond_2_1;
      if(_myram_cond_3_1) begin
        myram_1_wenable <= 0;
      end 
      if(_myram_cond_4_1) begin
        _tmp_4 <= 1;
      end 
      _myram_cond_5_2 <= _myram_cond_5_1;
      if(xfsm == 1) begin
        myram_0_addr <= xaddr;
        myram_0_wdata <= xcount;
        myram_0_wenable <= 1;
      end 
      _myram_cond_0_1 <= xfsm == 1;
      if((xfsm == 2) && (xaddr < 16)) begin
        myram_0_addr <= xaddr;
      end 
      _myram_cond_1_1 <= (xfsm == 2) && (xaddr < 16);
      _myram_cond_2_1 <= (xfsm == 2) && (xaddr < 16);
      if((yfsm == 1) && (yvalid && (yfsm == 1))) begin
        myram_1_addr <= yaddr;
        myram_1_wdata <= ydata;
        myram_1_wenable <= 1;
      end 
      _myram_cond_3_1 <= (yfsm == 1) && (yvalid && (yfsm == 1));
      if((yfsm == 2) && (yaddr < 16)) begin
        myram_1_addr <= yaddr;
      end 
      _myram_cond_4_1 <= (yfsm == 2) && (yaddr < 16);
      _myram_cond_5_1 <= (yfsm == 2) && (yaddr < 16);
    end
  end

  localparam xfsm_1 = 1;
  localparam xfsm_2 = 2;
  localparam xfsm_3 = 3;

  always @(posedge CLK) begin
    if(RST) begin
      xfsm <= xfsm_init;
      xaddr <= 0;
      xcount <= 0;
      _xaddr_1 <= 0;
      _xaddr_2 <= 0;
    end else begin
      _xaddr_1 <= xaddr;
      _xaddr_2 <= _xaddr_1;
      case(xfsm)
        xfsm_init: begin
          xaddr <= 0;
          xcount <= 0;
          xfsm <= xfsm_1;
        end
        xfsm_1: begin
          xaddr <= xaddr + 1;
          xcount <= xcount + 1;
          if(xcount == 15) begin
            xaddr <= 0;
            xcount <= 0;
          end 
          if(xcount == 15) begin
            xfsm <= xfsm_2;
          end 
        end
        xfsm_2: begin
          xaddr <= xaddr + 1;
          if(_tmp_1) begin
            $display("RAM0[%d] = %d", _xaddr_2, myram_0_rdata);
          end 
          if(_xaddr_1 == 16) begin
            xfsm <= xfsm_3;
          end 
        end
      endcase
    end
  end


  always @(posedge CLK) begin
    if(RST) begin
      _tmp_data_0 <= 0;
      _tmp_valid_0 <= 0;
      _tmp_2 <= 0;
      _tmp_3 <= 0;
      __dataflow_seq_0_cond_0_1 <= 0;
    end else begin
      if(__dataflow_seq_0_cond_0_1) begin
        _tmp_3 <= 0;
      end 
      if((_tmp_ready_0 || !_tmp_valid_0) && xready && xvalid) begin
        _tmp_data_0 <= xdata + 8'd100;
      end 
      if(_tmp_valid_0 && _tmp_ready_0) begin
        _tmp_valid_0 <= 0;
      end 
      if((_tmp_ready_0 || !_tmp_valid_0) && xready) begin
        _tmp_valid_0 <= xvalid;
      end 
      if(_tmp_1 && (xready || !_tmp_3)) begin
        _tmp_2 <= myram_0_rdata;
      end 
      if(_tmp_1 && (xready || !_tmp_3)) begin
        _tmp_3 <= 1;
      end 
      __dataflow_seq_0_cond_0_1 <= 1;
      if(_tmp_3 && !xready) begin
        _tmp_3 <= _tmp_3;
      end 
    end
  end

  localparam yfsm_1 = 1;
  localparam yfsm_2 = 2;

  always @(posedge CLK) begin
    if(RST) begin
      yfsm <= yfsm_init;
      yaddr <= 0;
      _yaddr_1 <= 0;
      _yaddr_2 <= 0;
    end else begin
      _yaddr_1 <= yaddr;
      _yaddr_2 <= _yaddr_1;
      case(yfsm)
        yfsm_init: begin
          yaddr <= 0;
          yfsm <= yfsm_1;
        end
        yfsm_1: begin
          if(yvalid && (yfsm == 1)) begin
            yaddr <= yaddr + 1;
          end 
          if(yaddr == 15) begin
            yaddr <= 0;
          end 
          if(yaddr == 15) begin
            yfsm <= yfsm_2;
          end 
        end
        yfsm_2: begin
          if(yaddr < 16) begin
            yaddr <= yaddr + 1;
          end 
          if(_tmp_4) begin
            $display("RAM1[%d] = %d", _yaddr_2, myram_1_rdata);
          end 
        end
      endcase
    end
  end


endmodule



module myram
(
  input CLK,
  input [14-1:0] myram_0_addr,
  output [32-1:0] myram_0_rdata,
  input [32-1:0] myram_0_wdata,
  input myram_0_wenable,
  input [14-1:0] myram_1_addr,
  output [32-1:0] myram_1_rdata,
  input [32-1:0] myram_1_wdata,
  input myram_1_wenable
);

  reg [14-1:0] myram_0_daddr;
  reg [14-1:0] myram_1_daddr;
  reg [32-1:0] mem [0:16384-1];

  always @(posedge CLK) begin
    if(myram_0_wenable) begin
      mem[myram_0_addr] <= myram_0_wdata;
    end 
    myram_0_daddr <= myram_0_addr;
  end

  assign myram_0_rdata = mem[myram_0_daddr];

  always @(posedge CLK) begin
    if(myram_1_wenable) begin
      mem[myram_1_addr] <= myram_1_wdata;
    end 
    myram_1_daddr <= myram_1_addr;
  end

  assign myram_1_rdata = mem[myram_1_daddr];

endmodule
"""


def test():
    veriloggen.reset()
    test_module = types_ram_manager_rtl_dataflow.mkTest()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import patternmux

expected_verilog = """
module blinkled #
(
  parameter WIDTH = 8
)
(
  input CLK,
  input RST,
  output reg [WIDTH-1:0] LED
);

  reg [32-1:0] count;

  always @(posedge CLK) begin
    if(RST) begin
      count <= 0;
      LED <= 0;
    end else begin
      count <= (count < 10)? count + 1 : 
               (count == 1023)? 0 : count + 1;
      LED <= (count < 10)? LED : 
             (count == 1023)? LED + 1 : LED;
    end
  end


endmodule
"""

def test():
    veriloggen.reset()
    test_module = patternmux.mkLed()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from __future__ import absolute_import
from __future__ import print_function
import sys
import os
import collections

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))))

from veriloggen import *

def mkLedTest():
    filename = os.path.dirname(os.path.abspath(__file__)) + '/led.v'
    modules = from_verilog.read_verilog_module(filename)
    test = modules['test']
    return test

if __name__ == '__main__':
    test = mkLedTest()
    verilog = test.to_verilog()
    print(verilog)



from __future__ import absolute_import
from __future__ import print_function
import sys
import os

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))))

from veriloggen import *

def mkLed():
    m = Module('blinkled')
    clk = m.Input('CLK')
    rst = m.Input('RST')
    
    x = m.Input('x', 32)
    vx = m.Input('vx')
    rx = m.Output('rx')
    
    y = m.Input('y', 32)
    vy = m.Input('vy')
    ry = m.Output('ry')

    z = m.Output('z', 32)
    vz = m.Output('vz')
    rz = m.Input('rz')
    
    df = Pipeline(m, 'df', clk, rst)
    
    px = df.input(x, valid=vx, ready=rx)
    py = df.input(y, valid=vy, ready=ry)
    pz = df(px + py)
    pz.output(z, valid=vz, ready=rz)
    
    df.make_always()

    try:
        df.draw_graph()
    except:
        print('Dataflow graph could not be generated.', file=sys.stderr)
    
    return m

def mkTest(numports=8):
    m = Module('test')

    # target instance
    led = mkLed()
    
    # copy paras and ports
    params = m.copy_params(led)
    ports = m.copy_sim_ports(led)

    clk = ports['CLK']
    rst = ports['RST']
    
    x = ports['x']
    vx = ports['vx']
    rx = ports['rx']
    y = ports['y']
    vy = ports['vy']
    ry = ports['ry']
    z = ports['z']
    vz = ports['vz']
    rz = ports['rz']
    
    uut = m.Instance(led, 'uut',
                     params=m.connect_params(led),
                     ports=m.connect_ports(led))

    reset_done = m.Reg('reset_done', initval=0)
    
    reset_stmt = []
    reset_stmt.append( reset_done(0) )
    reset_stmt.append( x(0) )
    reset_stmt.append( y(0) )
    reset_stmt.append( vx(0) )
    reset_stmt.append( vy(0) )
    reset_stmt.append( rz(0) )
    
    simulation.setup_waveform(m, uut)
    simulation.setup_clock(m, clk, hperiod=5)
    init = simulation.setup_reset(m, rst, reset_stmt, period=100)

    nclk = simulation.next_clock
    
    init.add(
        Delay(1000),
        reset_done(1),
        nclk(clk),
        Delay(10000),
        Systask('finish'),
    )

    
    x_count = m.TmpReg(32, initval=0)
    y_count = m.TmpReg(32, initval=0)
    z_count = m.TmpReg(32, initval=0)

    
    xfsm = FSM(m, 'xfsm', clk, rst)
    xfsm.add(vx(0))
    xfsm.goto_next(cond=reset_done)
    xfsm.add(vx(1))
    xfsm.add(x.inc(), cond=rx)
    xfsm.add(x_count.inc(), cond=rx)
    xfsm.goto_next(cond=AndList(x_count==10, rx))
    xfsm.add(vx(0))
    xfsm.make_always()
    
    
    yfsm = FSM(m, 'yfsm', clk, rst)
    yfsm.add(vy(0))
    yfsm.goto_next(cond=reset_done)
    yfsm.add(vy(1))
    yfsm.add(y.add(2), cond=ry)
    yfsm.add(y_count.inc(), cond=ry)
    yfsm.goto_next(cond=AndList(y_count==10, ry))
    yfsm.add(vy(0))
    yfsm.make_always()

    
    zfsm = FSM(m, 'zfsm', clk, rst)
    zfsm.add(rz(0))
    zfsm.goto_next(cond=reset_done)
    zfsm.goto_next()
    zinit= zfsm.current
    zfsm.add(rz(1), cond=vz)
    zfsm.goto_next(cond=vz)
    for i in range(10):
        zfsm.add(rz(0))
        zfsm.goto_next()
    zfsm.goto(zinit)
    zfsm.make_always()

    
    m.Always(Posedge(clk))(
        If(reset_done)(
            If(AndList(vx, rx))(
                Systask('display', 'x=%d', x)
            ),
            If(AndList(vy, ry))(
                Systask('display', 'y=%d', y)
            ),
            If(AndList(vz, rz))(
                Systask('display', 'z=%d', z)
            )
        )
    )
    
    return m
    
if __name__ == '__main__':
    test = mkTest()
    verilog = test.to_verilog('tmp.v')
    print(verilog)

    # run simulator (Icarus Verilog)
    sim = simulation.Simulator(test)
    rslt = sim.run() # display=False
    #rslt = sim.run(display=True)
    print(rslt)

    # launch waveform viewer (GTKwave)
    #sim.view_waveform() # background=False
    #sim.view_waveform(background=True)



from __future__ import absolute_import
from __future__ import print_function
import sys
import os

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(
    os.path.dirname(os.path.abspath(__file__)))))

from veriloggen import *
import veriloggen.types.axi as axi
import veriloggen.types.ram as ram
import veriloggen.dataflow as dataflow


# matrix-vecter multiplication
def mkMain():
    m = Module('main')
    clk = m.Input('CLK')
    rst = m.Input('RST')

    # AXI ports
    slave = axi.AxiSlave(m, 'slave', clk, rst)
    master = axi.AxiMaster(m, 'master', clk, rst)

    # a, b: source, c: result
    ram_a = ram.SyncRAMManager(m, 'ram_a', clk, rst, numports=1)
    ram_b = ram.SyncRAMManager(m, 'ram_b', clk, rst, numports=1)
    ram_c = ram.SyncRAMManager(m, 'ram_c', clk, rst, numports=1)

    read_fsm = FSM(m, 'read_fsm', clk, rst)
    write_fsm = FSM(m, 'write_fsm', clk, rst)
    df = dataflow.DataflowManager(m, clk, rst)
    # df.enable_draw_graph()

    read_fsm.goto_next()

    row_count = m.Reg('row_count', 32, initval=0)
    read_fsm(
        row_count(0)
    )

    # wait for slave request
    slave_addr, slave_counter, slave_valid = slave.pull_write_request(
        cond=read_fsm)
    read_fsm.If(slave_valid).goto_next()

    data, mask, valid, last = slave.pull_write_data(
        slave_counter, cond=read_fsm)
    read_fsm.If(valid).goto_next()

    write_fsm.If(read_fsm).goto_next()

    # computation
    master_addr = 1024 * 2
    ram_addr = 0
    length = 16
    dma_done = master.dma_read(
        ram_b, master_addr, ram_addr, length, cond=read_fsm)
    read_fsm.If(dma_done).goto_next()

    comp_start = read_fsm.current

    master_addr = 1024
    ram_addr = 0
    length = 16
    dma_done = master.dma_read(
        ram_a, master_addr, ram_addr, length, cond=read_fsm)
    read_fsm.If(dma_done).goto_next()

    adata, alast, adone = ram_a.read_dataflow(
        0, ram_addr, length, cond=read_fsm)
    bdata, blast, bdone = ram_b.read_dataflow(
        0, ram_addr, length, cond=read_fsm)
    read_fsm.goto_next()

    mul = adata * bdata
    mul_count = df.Counter(maxval=length)
    wcond = mul_count == 0
    cdata = df.Iadd(mul, reset=wcond.prev(1))

    read_fsm(
        row_count.inc()
    )
    read_fsm.If(row_count < length - 1).goto(comp_start)
    read_fsm.If(row_count == length - 1).goto_next()

    done = ram_c.write_dataflow(
        0, 0, cdata, length, cond=write_fsm, when=wcond)
    write_fsm.goto_next()

    write_fsm.If(done).goto_next()

    master_addr = 1024 * 3

    dma_done = master.dma_write(
        ram_c, master_addr, ram_addr, length, cond=write_fsm)
    write_fsm.If(dma_done).goto_next()

    read_fsm.If(write_fsm).goto_init()
    write_fsm.goto_init()

    seq = Seq(m, 'seq', clk, rst)
    seq.If(ram_c[0].wenable)(
        Systask('display', '[%d]<-%d', ram_c[0].addr, ram_c[0].wdata)
    )

    return m


def mkTest():
    m = Module('test')

    # target instance
    main = mkMain()

    # copy paras and ports
    params = m.copy_params(main)
    ports = m.copy_sim_ports(main)

    clk = ports['CLK']
    rst = ports['RST']

    #--------------------------------------------------------------------------
    # Master

    # awready (no stall)
    #awready = ports['master_awready']
    #_awready = m.TmpWireLike(awready)
    #_awready.assign(1)
    # m.Always()(awready(_awready))

    # wready (nostall)
    #wready = ports['master_wready']
    #_wready = m.TmpWireLike(wready)
    #_wready.assign(1)
    # m.Always()(wready(_wready))

    # awready (with stall)
    waddr_fsm = FSM(m, 'waddr', clk, rst)
    _awlen = m.Reg('_awlen', 32, initval=0)

    waddr_fsm(
        ports['master_awready'](0),
        ports['master_wready'](0),
        _awlen(0)
    )
    waddr_fsm.If(ports['master_awvalid']).goto_next()

    waddr_fsm.If(ports['master_awvalid'])(
        ports['master_awready'](1)
    )
    waddr_fsm.goto_next()

    waddr_fsm(
        ports['master_awready'](0),
        _awlen(ports['master_awlen'])
    )
    waddr_fsm.goto_next()

    # wready (with stall)
    waddr_init = waddr_fsm.current
    waddr_fsm(
        ports['master_wready'](0)
    )
    waddr_fsm.If(ports['master_wvalid']).goto_next()

    waddr_fsm.If(ports['master_wvalid'])(
        ports['master_wready'](1)
    )
    waddr_fsm.goto_next()

    waddr_fsm(
        ports['master_wready'](0)
    )
    waddr_fsm.goto_next()
    waddr_fsm.goto_next()
    waddr_fsm.goto_next()

    waddr_fsm(
        _awlen.dec()
    )
    waddr_fsm.goto(waddr_init)
    waddr_fsm.If(_awlen == 0).goto_init()

    # wready (no stall)
#    waddr_fsm(
#        ports['master_wready'](1)
#    )
#    waddr_fsm.Delay(1)(
#        ports['master_wready'](0)
#    )
#    waddr_fsm.If(ports['master_wvalid'])(
#        _awlen.dec()
#    )
#    waddr_fsm.Then().If(_awlen == 0).goto_next()

    # arready (no stall)
    #arready = ports['master_arready']
    #_arready = m.TmpWireLike(arready)
    #_arready.assign(0)
    #m.Always()( arready(_arready) )

    # arready, rvalid, rdata, rlast
    raddr_fsm = FSM(m, 'raddr', clk, rst)
    _arlen = m.Reg('_arlen', 32, initval=0)
    raddr_fsm(
        ports['master_arready'](0),
        ports['master_rdata'](-1),
        ports['master_rvalid'](0),
        ports['master_rlast'](0)
    )
    raddr_fsm.If(ports['master_arvalid']).goto_next()

    raddr_fsm.If(ports['master_arvalid'])(
        ports['master_arready'](1),
        ports['master_rdata'](ports['master_araddr'] - 1)
    )
    raddr_fsm.goto_next()

    raddr_fsm(
        ports['master_arready'](0),
        _arlen(ports['master_arlen'])
    )
    raddr_fsm.goto_next()

    ack = Ors(ports['master_rready'], Not(ports['master_rvalid']))

    # nodelay
#    raddr_fsm.If(Ands(ack, Not(ports['master_rlast'])))(
#        ports['master_rdata'].inc(),
#        ports['master_rvalid'](1),
#        ports['master_rlast'](0),
#        _arlen.dec()
#    )
#    raddr_fsm.Then().If(_arlen == 0)(
#        ports['master_rlast'](1),
#    )
#    raddr_fsm.Delay(1)(
#        ports['master_rvalid'](0),
#        ports['master_rlast'](0)
#    )
#    raddr_fsm.If(Ands(ports['master_rvalid'], Not(ports['master_rready'])))(
#        ports['master_rvalid'](ports['master_rvalid']),
#        ports['master_rlast'](ports['master_rlast']),
#    )
#    raddr_fsm.If(Ands(ports['master_rvalid'], ports[
#                 'master_rready'], ports['master_rlast'])).goto_next()

    rdata_head = raddr_fsm.current
    raddr_fsm.If(Ands(ack, Not(ports['master_rlast'])))(
        # ports['master_rdata'].inc(),
        ports['master_rdata'](1),
        ports['master_rvalid'](1),
        ports['master_rlast'](0),
        _arlen.dec()
    )
    raddr_fsm.Then().If(_arlen == 0)(
        ports['master_rlast'](1)
    )
    raddr_fsm.Delay(1)(
        ports['master_rvalid'](0),
        ports['master_rlast'](0)
    )
    raddr_fsm.goto_next()
    raddr_fsm.If(Ands(ports['master_rvalid'], Not(ports['master_rready'])))(
        ports['master_rvalid'](ports['master_rvalid']),
        ports['master_rlast'](ports['master_rlast']),
    )
    raddr_fsm.If(Ands(ports['master_rvalid'], ports[
                 'master_rready'])).goto_next()
    raddr_fsm.goto_next()
    raddr_fsm.goto_next()
    raddr_fsm.goto_next()
    raddr_fsm.If((_arlen + 1) & 0xff != 0).goto(rdata_head)
    raddr_fsm.If((_arlen + 1) & 0xff == 0).goto_next()

    raddr_fsm.goto_next()

    raddr_fsm.goto_init()

    #--------------------------------------------------------------------------
    # Slave
    awvalid = ports['slave_awvalid']
    _awvalid = m.TmpWireLike(awvalid)
    _awvalid.assign(1)  # on
    m.Always()(awvalid(_awvalid))

    awaddr = ports['slave_awaddr']
    _awaddr = m.TmpWireLike(awaddr)
    _awaddr.assign(0x100)  # on
    m.Always()(awaddr(_awaddr))

    awlen = ports['slave_awlen']
    _awlen = m.TmpWireLike(awlen)
    _awlen.assign(0)  # on
    m.Always()(awlen(_awlen))

    arvalid = ports['slave_arvalid']
    _arvalid = m.TmpWireLike(arvalid)
    _arvalid.assign(1)  # on
    m.Always()(arvalid(_arvalid))

    araddr = ports['slave_araddr']
    _araddr = m.TmpWireLike(araddr)
    _araddr.assign(0x100)  # on
    m.Always()(araddr(_araddr))

    arlen = ports['slave_arlen']
    _arlen = m.TmpWireLike(arlen)
    _arlen.assign(0)  # on
    m.Always()(arlen(_arlen))

    wvalid = ports['slave_wvalid']
    _wvalid = m.TmpWireLike(wvalid)
    _wvalid.assign(1)  # on
    m.Always()(wvalid(_wvalid))

    wdata = ports['slave_wdata']
    _wdata = m.TmpWireLike(wdata)
    _wdata.assign(0x200)  # on
    m.Always()(wdata(_wdata))

    wstrb = ports['slave_wstrb']
    _wstrb = m.TmpWireLike(wstrb)
    _wstrb.assign(Repeat(Int(1, 1), 32 // 8))  # on
    m.Always()(wstrb(_wstrb))

    wlast = ports['slave_wlast']
    _wlast = m.TmpWireLike(wlast)
    _wlast.assign(1)  # on
    m.Always()(wlast(_wlast))

    rready = ports['slave_rready']
    _rready = m.TmpWireLike(rready)
    _rready.assign(1)  # on
    m.Always()(rready(_rready))

    uut = m.Instance(main, 'uut',
                     params=m.connect_params(main),
                     ports=m.connect_ports(main))

    simulation.setup_waveform(m, uut, m.get_vars())
    simulation.setup_clock(m, clk, hperiod=5)
    init = simulation.setup_reset(m, rst, m.make_reset(), period=100)

    init.add(
        Delay(1000 * 100),
        Systask('finish'),
    )

    return m

if __name__ == '__main__':
    test = mkTest()
    verilog = test.to_verilog('tmp.v')
    print(verilog)

    sim = simulation.Simulator(test)
    rslt = sim.run()
    print(rslt)

    # sim.view_waveform()



from __future__ import absolute_import
from __future__ import print_function

import veriloggen.core.vtypes as vtypes
import veriloggen.core.module as module
from veriloggen.seq.seq import Seq
from . import util


class FifoWriteInterface(object):
    _I = 'Reg'
    _O = 'Wire'

    def __init__(self, m, name=None, datawidth=32, itype=None, otype=None,
                 p_enq='enq', p_wdata='wdata', p_full='full', p_almost_full='almost_full'):

        if itype is None:
            itype = self._I
        if otype is None:
            otype = self._O

        self.m = m

        name_enq = p_enq if name is None else '_'.join([name, p_enq])
        name_wdata = p_wdata if name is None else '_'.join([name, p_wdata])
        name_full = p_full if name is None else '_'.join([name, p_full])
        name_almost_full = p_almost_full if name is None else '_'.join(
            [name, p_almost_full])

        self.enq = util.make_port(m, itype, name_enq, initval=0)
        self.wdata = util.make_port(m, itype, name_wdata, datawidth, initval=0)
        self.full = util.make_port(m, otype, name_full, initval=0)
        self.almost_full = util.make_port(
            m, otype, name_almost_full, initval=0)

    def connect(self, targ):
        util.connect_port(self.enq, targ.enq)
        util.connect_port(self.wdata, targ.wdata)
        util.connect_port(targ.full, self.full)
        util.connect_port(targ.almost_full, self.almost_full)


class FifoReadInterface(object):
    _I = 'Reg'
    _O = 'Wire'

    def __init__(self, m, name=None, datawidth=32, itype=None, otype=None,
                 p_deq='deq', p_rdata='rdata', p_empty='empty', p_almost_empty='almost_empty'):

        if itype is None:
            itype = self._I
        if otype is None:
            otype = self._O

        self.m = m

        name_deq = p_deq if name is None else '_'.join([name, p_deq])
        name_rdata = p_rdata if name is None else '_'.join([name, p_rdata])
        name_empty = p_empty if name is None else '_'.join([name, p_empty])
        name_almost_empty = p_almost_empty if name is None else '_'.join(
            [name, p_almost_empty])

        self.deq = util.make_port(m, itype, name_deq, initval=0)
        self.rdata = util.make_port(m, otype, name_rdata, datawidth, initval=0)
        self.empty = util.make_port(m, otype, name_empty, initval=0)
        self.almost_empty = util.make_port(
            m, otype, name_almost_empty, initval=0)

    def connect(self, targ):
        util.connect_port(self.deq, targ.deq)
        util.connect_port(targ.rdata, self.rdata)
        util.connect_port(targ.empty, self.empty)
        util.connect_port(targ.almost_empty, self.almost_empty)


class FifoWriteSlaveInterface(FifoWriteInterface):
    _I = 'Input'
    _O = 'Output'


class FifoWriteMasterInterface(FifoWriteInterface):
    _I = 'Output'
    _O = 'Input'


class FifoReadSlaveInterface(FifoReadInterface):
    _I = 'Input'
    _O = 'Output'


class FifoReadMasterInterface(FifoReadInterface):
    _I = 'Output'
    _O = 'Input'


#-------------------------------------------------------------------------
def mkFifoDefinition(name, datawidth=32, addrwidth=4):
    m = module.Module(name)
    clk = m.Input('CLK')
    rst = m.Input('RST')

    wif = FifoWriteSlaveInterface(m, name, datawidth)
    rif = FifoReadSlaveInterface(m, name, datawidth)

    mem = m.Reg('mem', datawidth, length=2**addrwidth)
    head = m.Reg('head', addrwidth, initval=0)
    tail = m.Reg('tail', addrwidth, initval=0)

    is_empty = m.Wire('is_empty')
    is_almost_empty = m.Wire('is_almost_empty')
    is_full = m.Wire('is_full')
    is_almost_full = m.Wire('is_almost_full')

    mask = (2 ** addrwidth) - 1

    is_empty.assign(head == tail)
    is_almost_empty.assign(head == ((tail + 1) & mask))
    is_full.assign(((head + 1) & mask) == tail)
    is_almost_full.assign(((head + 2) & mask) == tail)

    rdata = m.Reg('rdata_reg', datawidth, initval=0)

    wif.full.assign(is_full)
    wif.almost_full.assign(vtypes.Ors(is_almost_full, is_full))
    rif.empty.assign(is_empty)
    rif.almost_empty.assign(vtypes.Ors(is_almost_empty, is_empty))

    seq = Seq(m, '', clk, rst)

    seq.If(vtypes.Ands(wif.enq, vtypes.Not(is_full)))(
        mem[head](wif.wdata),
        head.inc()
    )

    seq.If(vtypes.Ands(rif.deq, vtypes.Not(is_empty)))(
        rdata(mem[tail]),
        tail.inc()
    )

    rif.rdata.assign(rdata)

    seq.make_always()

    return m


#-------------------------------------------------------------------------
class Fifo(object):

    def __init__(self, m, name, clk, rst, datawidth=32, addrwidth=4):
        self.m = m
        self.name = name
        self.clk = clk
        self.rst = rst
        self.datawidth = datawidth
        self.addrwidth = addrwidth
        self.wif = FifoWriteInterface(self.m, name, datawidth)
        self.rif = FifoReadInterface(self.m, name, datawidth)
        self.definition = mkFifoDefinition(name, datawidth, addrwidth)
        self.inst = self.m.Instance(self.definition, 'inst_' + name,
                                    ports=m.connect_ports(self.definition))

        self.seq = Seq(m, name, clk, rst)
        # self.m.add_hook(self.seq.make_always)

        # entry counter
        self._max_size = (2 ** self.addrwidth - 1 if isinstance(self.addrwidth, int) else
                          vtypes.Int(2) ** self.addrwidth - 1)

        self._count = self.m.Reg(
            'count_' + name, self.addrwidth + 1, initval=0)

        self.seq.If(
            vtypes.Ands(vtypes.Ands(self.wif.enq, vtypes.Not(self.wif.full)),
                        vtypes.Ands(self.rif.deq, vtypes.Not(self.rif.empty))))(
            self._count(self._count)
        ).Elif(vtypes.Ands(self.wif.enq, vtypes.Not(self.wif.full)))(
            self._count.inc()
        ).Elif(vtypes.Ands(self.rif.deq, vtypes.Not(self.rif.empty)))(
            self._count.dec()
        )

        self._enq_disabled = False
        self._deq_disabled = False

    def disable_enq(self):
        self.seq(
            self.wif.enq(0)
        )
        self._enq_disabled = True

    def disable_deq(self):
        self.seq(
            self.rif.deq(0)
        )
        self._deq_disabled = True

    def enq(self, wdata, cond=None, delay=0):
        """ Enque operation """
        if self._enq_disabled:
            raise TypeError('Enq disabled.')

        if cond is not None:
            self.seq.If(cond)

        current_delay = self.seq.current_delay

        not_full = vtypes.Not(self.wif.full)
        ack = vtypes.Ands(not_full, self.wif.enq)
        if current_delay + delay == 0:
            ready = vtypes.Not(self.wif.almost_full)
        else:
            ready = self._count + (current_delay + delay + 1) < self._max_size

        self.seq.Delay(current_delay + delay).EagerVal().If(not_full)(
            self.wif.wdata(wdata)
        )
        self.seq.Then().Delay(current_delay + delay)(
            self.wif.enq(1)
        )

        # de-assert
        self.seq.Delay(current_delay + delay + 1)(
            self.wif.enq(0)
        )

        return ack, ready

    def deq(self, cond=None, delay=0):
        """ Deque operation """
        if self._deq_disabled:
            raise TypeError('Deq disabled.')

        if cond is not None:
            self.seq.If(cond)

        not_empty = vtypes.Not(self.rif.empty)

        current_delay = self.seq.current_delay

        self.seq.Delay(current_delay + delay)(
            self.rif.deq(1)
        )

        rdata = self.rif.rdata
        rvalid = self.m.TmpReg(initval=0)

        self.seq.Then().Delay(current_delay + delay + 1)(
            rvalid(vtypes.Ands(not_empty, self.rif.deq))
        )

        # de-assert
        self.seq.Delay(current_delay + delay + 1)(
            self.rif.deq(0)
        )
        self.seq.Delay(current_delay + delay + 2)(
            rvalid(0)
        )

        return rdata, rvalid

    @property
    def wdata(self):
        return self.wif.wdata

    @property
    def empty(self):
        return self.rif.empty

    @property
    def almost_empty(self):
        return self.rif.almost_empty

    @property
    def rdata(self):
        return self.rif.rdata

    @property
    def full(self):
        return self.wif.full

    @property
    def almost_full(self):
        return self.wif.almost_full

    @property
    def count(self):
        return self._count

    @property
    def space(self):
        if isinstance(self._max_size, int):
            return vtypes.Int(self._max_size) - self.count
        return self._max_size - self.count

    def has_space(self, num=1):
        if num < 1:
            return True
        return (self._count + num < self._max_size)



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import from_verilog_branchpredunit

expected_verilog = """
`default_nettype none

`define BP_ON  
`ifdef BP_ON  
  
//Gshare Branch Predict
module BP #
  (
    parameter W_A = 19,
    parameter W_BHR = 6,
    parameter W_PHT = 10,
    parameter N_PHT = 1024,
    parameter BHR_OFFSET = 4
  )
  (
    input CLK,
    input RST_X,
    input EN,
    input [W_A-1:0] pc,
    output pred,
    input [W_A-1:0] update_pc,
    input update,
    input rslt
  );

    //Branch History Register
    reg [W_BHR-1:0] bhr;
    //Pattern History Table
    wire [1:0] read_pht;
    wire [1:0] update_read_pht;
    wire [1:0] write_pht;

    reg [W_BHR-1:0] d_bhr;
    reg d_update;
    reg d_rslt;
    reg [W_A-1:0] d_update_pc;

    pht_mem #(.W_PHT(W_PHT), .N_PHT(N_PHT))
    pht_mem(CLK, EN, pht_index(bhr, pc), read_pht,
            pht_index(bhr, update_pc), update_read_pht,
            pht_index(d_bhr, d_update_pc), write_pht, d_update);
    
    assign write_pht = pht_update(update_read_pht, d_rslt);
    assign pred = get_pred(read_pht);
    
    always @(posedge CLK or negedge RST_X) begin
        if(!RST_X) begin
            d_update <= 0;
            d_rslt <= 0;
            d_update_pc <= 0;
            d_bhr <= 0;
            bhr <= 0;
        end else begin
          if(EN) begin
            if(update) begin
                bhr <= {bhr, rslt};
            end
            d_update <= update;
            d_rslt <= rslt;
            d_update_pc <= update_pc;
            d_bhr <= bhr;
          end
        end
    end
    
    function [W_PHT-1:0] pht_index;
        input [W_BHR-1:0] bhr;
        input [W_A-1:0] pc;
        begin
          pht_index = (bhr << BHR_OFFSET) ^ (pc >> 2);
        end 
    endfunction

    function [1:0] pht_update;
        input [1:0] current;
        input rslt;
        begin
            if(rslt) begin
              pht_update = (current < 3)? current + 1: 3;
            end else begin
              pht_update = (current > 0)? current - 1: 0;
            end
        end
    endfunction
    
    function [0:0] get_pred;
        input [1:0] cnt;
        begin
          get_pred = (cnt >= 2)? 1:0;
        end
    endfunction
endmodule

module pht_mem #
 (
    parameter W_PHT = 10,
    parameter N_PHT = 1024
 )
 (
    input CLK,
    input EN,
    input [W_PHT-1:0] I_A0,
    output [1:0] O_Q0,
    input [W_PHT-1:0] I_A1,
    output [1:0] O_Q1,
    input [W_PHT-1:0] I_A2,
    input [1:0] I_D2,
    input I_WE2
 );

    reg [1:0] mem [0:N_PHT-1];
    reg [W_PHT-1:0] d_I_A0;
    reg [W_PHT-1:0] d_I_A1;

    integer i;
    initial begin
        d_I_A0 = 0;
        d_I_A1 = 0;
        for(i=0; i<N_PHT; i=i+1) begin
            mem[i] = 2'h3;
        end
    end
    always @(posedge CLK) begin
      if(EN) begin
            d_I_A0 <= I_A0;
            d_I_A1 <= I_A1;
            if(I_WE2) begin mem[I_A2] <= I_D2; end
      end
    end
    assign O_Q0 = mem[d_I_A0];
    assign O_Q1 = mem[d_I_A1];
endmodule

module BTB #
 (
    parameter N_WAY = 2, //2-way
    parameter W_A = 19,
    //parameter W_TAB = 10, //per way
    //parameter N_TAB = 1024, //per way
    parameter W_TAB = 5, //per way
    parameter N_TAB = 32 //per way
 )
 (
    input CLK, RST_X,
    input EN,
    input [W_A-1:0] pc,
    output [W_A-1:0] pred_target,
    output exist,
    output read_ras,
    input [W_A-1:0] update_pc,
    input [W_A-1:0] update_target,
    input update_read_ras,
    input update,
    input rslt
 );

    localparam W_TAG = W_A - (W_TAB + 2);

    reg [W_A-1:0] d_pc;
    reg [W_A-1:0] d_update_pc;
    reg [W_A-1:0] d_update_target;
    reg d_update;
    reg d_rslt;
    reg d_update_read_ras;
    
//    reg [0:0] last [0:N_TAB-1]; //LRU
//    reg [0:0] d_last;
    
    function [W_TAB-1:0] btb_index;
        input [W_A-1:0] pc;
        begin
          btb_index = pc >> 2;
        end
    endfunction
    
    function [W_TAG-1:0] get_tag;
        input [W_A-1:0] pc;
        begin
           get_tag = pc[W_A-1:(W_A-W_TAG)];
        end
    endfunction

/* -----\/----- EXCLUDED -----\/-----
    function lru;
        input last;
        input way;
        begin
          lru = ~(last == way);
        end
    endfunction
 -----/\----- EXCLUDED -----/\----- */

//    genvar w;
//    generate for(w=0;w<N_WAY;w=w+1) begin:way
        wire read_hit;
        wire update_hit;
        wire victim;
        reg [W_A-1:0] btb [0:N_TAB-1];
        reg [W_TAG-1:0] tag [0:N_TAB-1];
        wire [W_TAG-1:0] read_tag;
        wire [W_A-1:0] read_target;
        wire [W_TAG-1:0] update_tag;
        wire t_read_ras;
        reg [0:0] ras [0:N_TAB-1];
        wire btb_we;
        assign btb_we = d_rslt && d_update && (update_hit || victim);
        btb_mem #(.W_A(W_A), .W_TAB(W_TAB), .N_TAB(N_TAB))
        btb_mem (CLK, EN, btb_index(pc), read_target, 
                 btb_index(d_update_pc), d_update_target, btb_we);
        btb_tag #(.W_A(W_A), .W_TAB(W_TAB), .N_TAB(N_TAB))
        btb_tag (CLK, EN, btb_index(pc), read_tag, 
                 btb_index(update_pc), update_tag, 
                 btb_index(d_update_pc), get_tag(d_update_pc), btb_we);
        read_ras_mem #(.W_TAB(W_TAB), .N_TAB(N_TAB))
        ras_mem (CLK, EN, btb_index(pc), t_read_ras,
                 btb_index(d_update_pc), d_update_read_ras, btb_we);
        assign read_hit = get_tag(d_pc) == read_tag;
//    end endgenerate
  
/* -----\/----- EXCLUDED -----\/-----
    assign exist        = way[0].read_hit || way[1].read_hit;
    assign pred_target  = way[1].read_hit? way[1].read_target:
                         way[0].read_hit? way[0].read_target:
                         0;
    assign read_ras = way[1].read_hit? way[1].t_read_ras:
                      way[0].read_hit? way[0].t_read_ras:
                      0;
 -----/\----- EXCLUDED -----/\----- */
    assign exist        = read_hit;
    assign pred_target  = read_hit? read_target: 0;
    assign read_ras     = read_hit? t_read_ras: 0;
  
    integer j;
    always @(posedge CLK or negedge RST_X) begin
        if(!RST_X) begin
/* -----\/----- EXCLUDED -----\/-----
            for(j=0;j<N_TAB;j=j+1) begin
                last[j] <= 0;
            end
 -----/\----- EXCLUDED -----/\----- */
            d_pc <= 0;
            d_update_pc <= 0;
/* -----\/----- EXCLUDED -----\/-----
            d_last <= 0;
 -----/\----- EXCLUDED -----/\----- */
            d_update <= 0;
            d_rslt <= 0;
            d_update_read_ras <= 0;
        end else begin
          if(EN) begin
            d_pc <= pc;
            d_update_pc <= update_pc;
/* -----\/----- EXCLUDED -----\/-----
            d_last <= last[btb_index(update_pc)];
 -----/\----- EXCLUDED -----/\----- */
            d_update <= update;
            d_update_target <= update_target;
            d_rslt <= rslt;
            d_update_read_ras <= update_read_ras;
/* -----\/----- EXCLUDED -----\/-----
            if(d_rslt && d_update) begin
                if(way[0].update_hit)
                  last[btb_index(d_update_pc)] <= 0;
                else if(way[1].update_hit)
                  last[btb_index(d_update_pc)] <= 1;
            end
            if(way[0].read_hit) begin
              last[btb_index(d_pc)] <= 0;
            end 
            else begin if(way[1].read_hit) begin
              last[btb_index(d_pc)] <= 1;
            end
 -----/\----- EXCLUDED -----/\----- */
           end
        end
    end
    
/* -----\/----- EXCLUDED -----\/-----
    assign way[0].update_hit = (get_tag(d_update_pc) == way[0].update_tag);
    assign way[1].update_hit = (get_tag(d_update_pc) == way[1].update_tag) && ~way[0].update_hit;
    assign way[0].victim = ~way[0].update_hit && ~way[1].update_hit && 
                           lru(d_last, 0);
    assign way[1].victim = ~way[0].update_hit && ~way[1].update_hit && 
                           lru(d_last, 1);
 -----/\----- EXCLUDED -----/\----- */
    assign update_hit = (get_tag(d_update_pc) == update_tag);
    assign victim = ~update_hit;
endmodule

module btb_mem #
 (
    parameter W_A = 19,
    parameter W_TAB = 5, //per way
    parameter N_TAB = 32 //per way
 )
 (
    input CLK,
    input EN,
    input [W_TAB-1:0] I_A0,
    output [W_A-1:0] O_Q0,
    input [W_TAB-1:0] I_A1,
    input [W_A-1:0] I_D1,
    input I_WE1
  );  

    reg [W_A-1:0] btb [0:N_TAB-1];
    reg [W_TAB-1:0] d_I_A0;

    integer i;
    initial begin
        d_I_A0 = 0;
        for(i=0; i<N_TAB; i=i+1) begin
            btb[i] = 0;
        end
    end
    always @(posedge CLK) begin
      if(EN) begin
          d_I_A0 <= I_A0;
          if(I_WE1) begin btb[I_A1] <= I_D1; end
      end
    end
    assign O_Q0 = btb[d_I_A0];
endmodule

module btb_tag #
 (
    parameter W_A = 19,
    parameter W_TAB = 5, //per way
    parameter N_TAB = 32 //per way
 )
 (
    input CLK,
    input EN,
    input [W_TAB-1:0] I_A0,
    output [W_TAG-1:0] O_Q0,
    input [W_TAB-1:0] I_A1,
    output [W_TAG-1:0] O_Q1,
    input [W_TAB-1:0] I_A2,
    input [W_TAG-1:0] I_D2,
    input I_WE2
  );
    
    localparam W_TAG = W_A - (W_TAB + 2);
    reg [W_TAG-1:0] tag [0:N_TAB-1];
    reg [W_TAB-1:0] d_I_A0;
    reg [W_TAB-1:0] d_I_A1;

    integer i;
    initial begin
        d_I_A0 = 0;
        d_I_A1 = 0;
        for(i=0; i<N_TAB; i=i+1) begin
            tag[i] = ~0;
        end
    end
    
    always @(posedge CLK) begin
      if(EN) begin
            d_I_A0 <= I_A0;
            d_I_A1 <= I_A1;
            if(I_WE2) begin tag[I_A2] <= I_D2; end
      end
    end
    assign O_Q0 = tag[d_I_A0];
    assign O_Q1 = tag[d_I_A1];
endmodule

module read_ras_mem #
  (
    parameter W_TAB = 5, //per way
    parameter N_TAB = 32 //per way
  )
  (
    input CLK,
    input EN,
    input [W_TAB-1:0] I_A0,
    output [0:0] O_Q0,
    input [W_TAB-1:0] I_A1,
    input [0:0] I_D1,
    input I_WE1
  );

    reg [0:0] ras [0:N_TAB-1];
    reg [W_TAB-1:0] d_I_A0;
    
    integer i;
    initial begin
        d_I_A0 = 0;
        for(i=0; i<N_TAB; i=i+1) begin
            ras[i] = 0;
        end
    end
    
    always @(posedge CLK) begin
      if(EN) begin
            d_I_A0 <= I_A0;
            if(I_WE1) begin ras[I_A1] <= I_D1; end
      end
    end
    assign O_Q0 = ras[d_I_A0];
endmodule

//Return Address Stack
module RAS #
  (
    parameter W_A = 19,
    parameter W_STACK = 5,
    parameter N_STACK = 32,
    parameter W_CNT = 32
  )
  (
    input CLK,
    input RST_X,
    input EN,
    input I_PUSH,
    input I_POP,
    input [W_A-1:0] I_ADDR,
    output reg [W_A-1:0] O_ADDR,
    output O_OVERFLOW
  );

    reg [W_A-1:0] stack [0:N_STACK];
    reg [W_STACK-1:0] head;
    reg [W_CNT-1:0] over_cnt;
    
    assign O_OVERFLOW = (over_cnt > 0);
    wire [W_STACK-1:0] read_ptr;
    assign read_ptr = (head == 0)? 0 : head -1;
    //assign O_ADDR = (head == 0)? 0: stack[read_ptr];
    
    always @(posedge CLK or negedge RST_X) begin
        if(!RST_X) begin
            head <= 0;
            over_cnt <= 0;
        end else begin
          if(EN) begin
            O_ADDR <= (head == 0)? 0: stack[read_ptr];
            if(I_PUSH && head != N_STACK-1) begin
                stack[head] <= I_ADDR;
                head <= head +1;
            end
            if(I_POP && head != 0) begin
                if(over_cnt == 0) begin head <= head -1; end
            end
            if(I_PUSH && head == N_STACK-1) begin
                over_cnt <= over_cnt +1;
            end
            if(I_POP && head == N_STACK-1) begin
                over_cnt <= over_cnt -1;
            end
          end
        end
    end
endmodule

`else // !`ifdef BP_ON

module BP #
 (
    parameter W_A = 19,
    parameter W_BHR = 6,
    parameter W_PHT = 10,
    parameter N_PHT = 1024,
    parameter BHR_OFFSET = 4
 )    
    input CLK, RST_X;
    input EN;
    input [W_A-1:0] pc;
    output pred;
    input [W_A-1:0] update_pc;
    input update;
    input rslt;
    assign pred = 0;
endmodule



module BTB #
 (
    parameter N_WAY = 2, //2-way
    parameter W_A = 19,
    //parameter W_TAB = 10, //per way
    //parameter N_TAB = 1024, //per way
    parameter W_TAB = 5, //per way
    parameter N_TAB = 32 //per way
 )
 (
    input CLK, RST_X,
    input EN,
    input [W_A-1:0] pc,
    output [W_A-1:0] pred_target,
    output exist,
    output read_ras,
    input [W_A-1:0] update_pc,
    input [W_A-1:0] update_target,
    input update_read_ras,
    input update,
    input rslt
 );
    localparam W_TAG = W_A - (W_TAB + 2);
    assign pred_target = pc + 4;
    assign exist = 0;
    assign read_ras = 0;
endmodule

module RAS #
  (
    parameter W_A = 19,
    parameter W_STACK = 5,
    parameter N_STACK = 32,
    parameter W_CNT = 32
  )
  (
    input CLK,
    input RST_X,
    input EN,
    input I_PUSH,
    input I_POP,
    input [W_A-1:0] I_ADDR,
    output [W_A-1:0] O_ADDR,
    output O_OVERFLOW
  );
    assign O_ADDR = 0;
    assign O_OVERFLOW = 0;
endmodule
`endif

`default_nettype wire
"""

def test():
    veriloggen.reset()
    test_modules = from_verilog_branchpredunit.mkMips()
    code = ''.join([ m.to_verilog() for m in test_modules.values() if not m.used ])

    from pyverilog.vparser.parser import parse
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    import sys
    import tempfile
    
    # encoding: 'utf-8' ?
    encode = sys.getdefaultencoding()
    
    tmp = tempfile.NamedTemporaryFile()
    tmp.write(expected_verilog.encode(encode))
    tmp.read()
    filename = tmp.name
    print(filename)
    
    expected_ast, _ = parse([ filename ])
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import dataflow_ireg

expected_verilog = """
module test;

  reg CLK;
  reg RST;
  reg [32-1:0] xdata;
  reg xvalid;
  wire xready;
  reg [1-1:0] resetdata;
  reg resetvalid;
  wire resetready;
  reg [1-1:0] enabledata;
  reg enablevalid;
  wire enableready;
  wire [32-1:0] zdata;
  wire zvalid;
  reg zready;

  main
  uut
  (
    .CLK(CLK),
    .RST(RST),
    .xdata(xdata),
    .xvalid(xvalid),
    .xready(xready),
    .resetdata(resetdata),
    .resetvalid(resetvalid),
    .resetready(resetready),
    .enabledata(enabledata),
    .enablevalid(enablevalid),
    .enableready(enableready),
    .zdata(zdata),
    .zvalid(zvalid),
    .zready(zready)
  );

  reg reset_done;

  initial begin
    $dumpfile("uut.vcd");
    $dumpvars(0, uut);
  end


  initial begin
    CLK = 0;
    forever begin
      #5 CLK = !CLK;
    end
  end


  initial begin
    RST = 0;
    reset_done = 0;
    xdata = 0;
    xvalid = 0;
    resetdata = 0;
    resetvalid = 0;
    zready = 0;
    #100;
    RST = 1;
    #100;
    RST = 0;
    #1000;
    reset_done = 1;
    @(posedge CLK);
    #1;
    #10000;
    $finish;
  end

  reg [32-1:0] xfsm;
  localparam xfsm_init = 0;
  reg [32-1:0] _tmp_0;
  localparam xfsm_1 = 1;
  localparam xfsm_2 = 2;
  localparam xfsm_3 = 3;
  localparam xfsm_4 = 4;
  localparam xfsm_5 = 5;
  localparam xfsm_6 = 6;
  localparam xfsm_7 = 7;
  localparam xfsm_8 = 8;
  localparam xfsm_9 = 9;
  localparam xfsm_10 = 10;
  localparam xfsm_11 = 11;
  localparam xfsm_12 = 12;
  localparam xfsm_13 = 13;
  localparam xfsm_14 = 14;
  localparam xfsm_15 = 15;
  localparam xfsm_16 = 16;
  localparam xfsm_17 = 17;
  localparam xfsm_18 = 18;
  localparam xfsm_19 = 19;
  localparam xfsm_20 = 20;
  localparam xfsm_21 = 21;
  localparam xfsm_22 = 22;
  localparam xfsm_23 = 23;
  localparam xfsm_24 = 24;

  always @(posedge CLK) begin
    if(RST) begin
      xfsm <= xfsm_init;
      _tmp_0 <= 0;
    end else begin
      case(xfsm)
        xfsm_init: begin
          xvalid <= 0;
          if(reset_done) begin
            xfsm <= xfsm_1;
          end 
        end
        xfsm_1: begin
          xfsm <= xfsm_2;
        end
        xfsm_2: begin
          xfsm <= xfsm_3;
        end
        xfsm_3: begin
          xfsm <= xfsm_4;
        end
        xfsm_4: begin
          xfsm <= xfsm_5;
        end
        xfsm_5: begin
          xfsm <= xfsm_6;
        end
        xfsm_6: begin
          xfsm <= xfsm_7;
        end
        xfsm_7: begin
          xfsm <= xfsm_8;
        end
        xfsm_8: begin
          xfsm <= xfsm_9;
        end
        xfsm_9: begin
          xfsm <= xfsm_10;
        end
        xfsm_10: begin
          xfsm <= xfsm_11;
        end
        xfsm_11: begin
          xvalid <= 1;
          xfsm <= xfsm_12;
        end
        xfsm_12: begin
          if(xready) begin
            xdata <= xdata + 1;
          end 
          if(xready) begin
            _tmp_0 <= _tmp_0 + 1;
          end 
          if((_tmp_0 == 5) && xready) begin
            xvalid <= 0;
          end 
          if((_tmp_0 == 5) && xready) begin
            xfsm <= xfsm_13;
          end 
        end
        xfsm_13: begin
          xfsm <= xfsm_14;
        end
        xfsm_14: begin
          xfsm <= xfsm_15;
        end
        xfsm_15: begin
          xfsm <= xfsm_16;
        end
        xfsm_16: begin
          xfsm <= xfsm_17;
        end
        xfsm_17: begin
          xfsm <= xfsm_18;
        end
        xfsm_18: begin
          xfsm <= xfsm_19;
        end
        xfsm_19: begin
          xfsm <= xfsm_20;
        end
        xfsm_20: begin
          xfsm <= xfsm_21;
        end
        xfsm_21: begin
          xfsm <= xfsm_22;
        end
        xfsm_22: begin
          xfsm <= xfsm_23;
        end
        xfsm_23: begin
          xvalid <= 1;
          if(xready) begin
            xdata <= xdata + 1;
          end 
          if(xready) begin
            _tmp_0 <= _tmp_0 + 1;
          end 
          if((_tmp_0 == 20) && xready) begin
            xvalid <= 0;
          end 
          if((_tmp_0 == 20) && xready) begin
            xfsm <= xfsm_24;
          end 
        end
      endcase
    end
  end

  reg [32-1:0] zfsm;
  localparam zfsm_init = 0;
  localparam zfsm_1 = 1;
  localparam zfsm_2 = 2;
  localparam zfsm_3 = 3;
  localparam zfsm_4 = 4;
  localparam zfsm_5 = 5;
  localparam zfsm_6 = 6;
  localparam zfsm_7 = 7;
  localparam zfsm_8 = 8;

  always @(posedge CLK) begin
    if(RST) begin
      zfsm <= zfsm_init;
    end else begin
      case(zfsm)
        zfsm_init: begin
          zready <= 0;
          if(reset_done) begin
            zfsm <= zfsm_1;
          end 
        end
        zfsm_1: begin
          zfsm <= zfsm_2;
        end
        zfsm_2: begin
          if(zvalid) begin
            zready <= 1;
          end 
          if(zvalid) begin
            zfsm <= zfsm_3;
          end 
        end
        zfsm_3: begin
          zready <= 0;
          zfsm <= zfsm_4;
        end
        zfsm_4: begin
          zready <= 0;
          zfsm <= zfsm_5;
        end
        zfsm_5: begin
          zready <= 0;
          zfsm <= zfsm_6;
        end
        zfsm_6: begin
          zready <= 0;
          zfsm <= zfsm_7;
        end
        zfsm_7: begin
          zready <= 0;
          zfsm <= zfsm_8;
        end
        zfsm_8: begin
          zfsm <= zfsm_2;
        end
      endcase
    end
  end

  reg [32-1:0] enable;
  localparam enable_init = 0;
  reg [32-1:0] enable_count;
  localparam enable_1 = 1;
  localparam enable_2 = 2;

  always @(posedge CLK) begin
    if(RST) begin
      enable <= enable_init;
      enable_count <= 0;
    end else begin
      case(enable)
        enable_init: begin
          if(reset_done) begin
            enable <= enable_1;
          end 
        end
        enable_1: begin
          enablevalid <= 1;
          if(enablevalid && enableready) begin
            enable_count <= enable_count + 1;
          end 
          if(enablevalid && enableready && (enable_count == 2)) begin
            enabledata <= 1;
          end 
          if(enablevalid && enableready && (enable_count == 2)) begin
            enable <= enable_2;
          end 
        end
        enable_2: begin
          if(enablevalid && enableready) begin
            enabledata <= 0;
          end 
          enable_count <= 0;
          if(enablevalid && enableready) begin
            enable <= enable_1;
          end 
        end
      endcase
    end
  end

  reg [32-1:0] reset;
  localparam reset_init = 0;
  reg [32-1:0] reset_count;
  localparam reset_1 = 1;
  localparam reset_2 = 2;

  always @(posedge CLK) begin
    if(RST) begin
      reset <= reset_init;
      reset_count <= 0;
    end else begin
      case(reset)
        reset_init: begin
          if(reset_done) begin
            reset <= reset_1;
          end 
        end
        reset_1: begin
          resetvalid <= 1;
          if(resetvalid && resetready) begin
            reset_count <= reset_count + 1;
          end 
          if(resetvalid && resetready && (reset_count == 2)) begin
            resetdata <= 0;
          end 
          if(resetvalid && resetready && (reset_count == 2)) begin
            reset <= reset_2;
          end 
        end
        reset_2: begin
          if(resetvalid && resetready) begin
            resetdata <= 0;
          end 
          reset_count <= 0;
          if(resetvalid && resetready) begin
            reset <= reset_1;
          end 
        end
      endcase
    end
  end


  always @(posedge CLK) begin
    if(reset_done) begin
      if(xvalid && xready) begin
        $display("xdata=%d", xdata);
      end 
      if(zvalid && zready) begin
        $display("zdata=%d", zdata);
      end 
    end 
  end


endmodule



module main
(
  input CLK,
  input RST,
  input [32-1:0] xdata,
  input xvalid,
  output xready,
  input [1-1:0] resetdata,
  input resetvalid,
  output resetready,
  input [1-1:0] enabledata,
  input enablevalid,
  output enableready,
  output [32-1:0] zdata,
  output zvalid,
  input zready
);

  reg [32-1:0] _tmp_data_0;
  reg _tmp_valid_0;
  wire _tmp_ready_0;
  assign enableready = (_tmp_ready_0 || !_tmp_valid_0) && (xvalid && enablevalid && resetvalid);
  assign xready = (_tmp_ready_0 || !_tmp_valid_0) && (xvalid && enablevalid && resetvalid);
  assign resetready = (_tmp_ready_0 || !_tmp_valid_0) && (xvalid && enablevalid && resetvalid);
  assign zdata = _tmp_data_0;
  assign zvalid = _tmp_valid_0;
  assign _tmp_ready_0 = zready;

  always @(posedge CLK) begin
    if(RST) begin
      _tmp_data_0 <= 1'd0;
      _tmp_valid_0 <= 0;
    end else begin
      if((_tmp_ready_0 || !_tmp_valid_0) && (xready && enableready && resetready) && (xvalid && enablevalid && resetvalid) && enabledata) begin
        _tmp_data_0 <= xdata;
      end 
      if(_tmp_valid_0 && _tmp_ready_0) begin
        _tmp_valid_0 <= 0;
      end 
      if((_tmp_ready_0 || !_tmp_valid_0) && (xready && enableready && resetready)) begin
        _tmp_valid_0 <= xvalid && enablevalid && resetvalid;
      end 
      if((_tmp_ready_0 || !_tmp_valid_0) && (xready && enableready && resetready) && (xvalid && enablevalid && resetvalid) && resetdata) begin
        _tmp_data_0 <= 1'd0;
      end 
      if((_tmp_ready_0 || !_tmp_valid_0) && (xready && enableready && resetready) && (xvalid && enablevalid && resetvalid) && enabledata && resetdata) begin
        _tmp_data_0 <= 1'd0;
      end 
    end
  end


endmodule
"""

def test():
    veriloggen.reset()
    test_module = dataflow_ireg.mkTest()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from __future__ import absolute_import
from __future__ import print_function
import sys
import os

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))))

from veriloggen import *
import veriloggen.dataflow as dataflow

def mkMain():
    m = Module('main')

    clk = m.Input('CLK')
    rst = m.Input('RST')

    df = dataflow.DataflowManager(m, clk, rst)
    
    x = df.Counter()
    y = df.Counter()
    z = x * 2 - y

    z.output('zdata', 'zvalid')

    return m

def mkTest(numports=8):
    m = Module('test')

    # target instance
    main = mkMain()

    params = m.copy_params(main)
    ports = m.copy_sim_ports(main)

    clk = ports['CLK']
    rst = ports['RST']

    uut = m.Instance(main, 'uut',
                     params=m.connect_params(main),
                     ports=m.connect_ports(main))

    reset_done = m.Reg('reset_done', initval=0)
    reset_stmt = []
    reset_stmt.append( reset_done(0) )

    simulation.setup_waveform(m, uut)
    simulation.setup_clock(m, clk, hperiod=5)
    init = simulation.setup_reset(m, rst, reset_stmt, period=100)

    nclk = simulation.next_clock
    
    init.add(
        Delay(1000),
        reset_done(1),
        nclk(clk),
        Delay(10000),
        Systask('finish'),
    )

    return m

    
if __name__ == '__main__':
    test = mkTest()
    verilog = test.to_verilog('tmp.v')
    print(verilog)

    # run simulator (Icarus Verilog)
    sim = simulation.Simulator(test)
    rslt = sim.run() # display=False
    #rslt = sim.run(display=True)
    print(rslt)

    # launch waveform viewer (GTKwave)
    #sim.view_waveform() # background=False
    #sim.view_waveform(background=True)



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import dataflow_counter

expected_verilog = """
module test;

  reg CLK;
  reg RST;
  wire [8-1:0] cdata;
  wire cvalid;

  main
  uut
  (
    .CLK(CLK),
    .RST(RST),
    .cdata(cdata),
    .cvalid(cvalid)
  );


  initial begin
    $dumpfile("uut.vcd");
    $dumpvars(0, uut, CLK, RST, cdata, cvalid);
  end


  initial begin
    CLK = 0;
    forever begin
      #5 CLK = !CLK;
    end
  end


  initial begin
    RST = 0;
    #100;
    RST = 1;
    #100;
    RST = 0;
    #100000;
    $finish;
  end


endmodule



module main
(
  input CLK,
  input RST,
  output [8-1:0] cdata,
  output cvalid
);

  reg [8-1:0] _tmp_data_0;
  reg _tmp_valid_0;
  wire _tmp_ready_0;
  assign _tmp_ready_0 = (_tmp_ready_2 || !_tmp_valid_2) && (_tmp_valid_0 && _tmp_valid_1);
  reg [8-1:0] _tmp_data_1;
  reg _tmp_valid_1;
  wire _tmp_ready_1;
  assign _tmp_ready_1 = (_tmp_ready_2 || !_tmp_valid_2) && (_tmp_valid_0 && _tmp_valid_1);
  reg [8-1:0] _tmp_data_2;
  reg _tmp_valid_2;
  wire _tmp_ready_2;
  assign cdata = _tmp_data_2;
  assign cvalid = _tmp_valid_2;
  assign _tmp_ready_2 = 1;
  reg [32-1:0] fsm;
  localparam fsm_init = 0;
  reg [32-1:0] _tmp_3;
  localparam fsm_1 = 1;

  always @(posedge CLK) begin
    if(RST) begin
      fsm <= fsm_init;
      _tmp_3 <= 0;
    end else begin
      case(fsm)
        fsm_init: begin
          if(cvalid) begin
            _tmp_3 <= _tmp_3 + 1;
            $display("c=%d", cdata);
          end 
          if(_tmp_3 == 32) begin
            fsm <= fsm_1;
          end 
        end
      endcase
    end
  end


  always @(posedge CLK) begin
    if(RST) begin
      _tmp_data_0 <= 1'd0;
      _tmp_valid_0 <= 0;
      _tmp_data_1 <= 1'd0;
      _tmp_valid_1 <= 0;
      _tmp_data_2 <= 0;
      _tmp_valid_2 <= 0;
    end else begin
      if((_tmp_ready_0 || !_tmp_valid_0) && 1 && 1) begin
        _tmp_data_0 <= (_tmp_data_0 >= 7)? 0 : _tmp_data_0 + 2'd1;
      end 
      if(_tmp_valid_0 && _tmp_ready_0) begin
        _tmp_valid_0 <= 0;
      end 
      if((_tmp_ready_0 || !_tmp_valid_0) && 1) begin
        _tmp_valid_0 <= 1;
      end 
      if((_tmp_ready_1 || !_tmp_valid_1) && 1 && 1) begin
        _tmp_data_1 <= (_tmp_data_1 >= 14)? 0 : _tmp_data_1 + 3'd2;
      end 
      if(_tmp_valid_1 && _tmp_ready_1) begin
        _tmp_valid_1 <= 0;
      end 
      if((_tmp_ready_1 || !_tmp_valid_1) && 1) begin
        _tmp_valid_1 <= 1;
      end 
      if((_tmp_ready_2 || !_tmp_valid_2) && (_tmp_ready_0 && _tmp_ready_1) && (_tmp_valid_0 && _tmp_valid_1)) begin
        _tmp_data_2 <= _tmp_data_0 + _tmp_data_1;
      end 
      if(_tmp_valid_2 && _tmp_ready_2) begin
        _tmp_valid_2 <= 0;
      end 
      if((_tmp_ready_2 || !_tmp_valid_2) && (_tmp_ready_0 && _tmp_ready_1)) begin
        _tmp_valid_2 <= _tmp_valid_0 && _tmp_valid_1;
      end 
    end
  end


endmodule
"""

def test():
    veriloggen.reset()
    test_module = dataflow_counter.mkTest()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from __future__ import absolute_import
from __future__ import print_function

from .thread import Thread, reset
from .pool import ThreadPool
from .ttypes import Mutex, Lock, Barrier, Shared, RAM, FIFO, AXIM, AXIS
from .stream import Stream
